"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from collections import deque
from typing import Any

from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.helpers import get_value_from_path
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword


class DistributionTargetOriginId(CfnLintKeyword):
    id = "E3057"
    shortdesc = "Validate that CloudFront TargetOriginId is a specified Origin"
    description = (
        "CloudFront TargetOriginId has to map to an Origin Id that "
        "is in the same DistributionConfig"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-cloudfront-distribution-defaultcachebehavior.html#cfn-cloudfront-distribution-defaultcachebehavior-targetoriginid"
    tags = ["properties", "cloudfront"]

    def __init__(self):
        """Init"""
        super().__init__(
            keywords=[
                "Resources/AWS::CloudFront::Distribution/Properties/DistributionConfig"
            ]
        )

    def validate(
        self, validator: Validator, _, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:

        for cache_origin_id, cache_validator in get_value_from_path(
            validator, instance, path=deque(["DefaultCacheBehavior", "TargetOriginId"])
        ):
            if not validator.is_type(cache_origin_id, "string"):
                continue
            origin_ids = []
            for origin_id, _ in get_value_from_path(
                cache_validator, instance, path=deque(["Origins", "*", "Id"])
            ):
                if not validator.is_type(origin_id, "string"):
                    break
                if origin_id == cache_origin_id:
                    break
                origin_ids.append(origin_id)
            else:
                for origin_id, _ in get_value_from_path(
                    cache_validator,
                    instance,
                    path=deque(["OriginGroups", "Items", "*", "Id"]),
                ):
                    if origin_id is None:
                        continue
                    if not validator.is_type(origin_id, "string"):
                        break
                    if origin_id == cache_origin_id:
                        break
                    origin_ids.append(origin_id)
                else:
                    yield ValidationError(
                        message=f"{cache_origin_id!r} is not one of {origin_ids!r}",
                        rule=self,
                        path_override=cache_validator.context.path.path,
                        validator="enum",
                    )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import logging
from collections import deque
from typing import Any

from cfnlint.data import AdditionalSpecs
from cfnlint.helpers import ensure_list, is_function, load_resource
from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.helpers import get_value_from_path
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword

LOGGER = logging.getLogger(__name__)


class _Arn:

    def __init__(self, full_arn: str):
        arn_parts = full_arn.split(":", 5)

        self._parts: list[str] = ["*", "*", "*", "*", "*", "*"]

        for i, arn_part in enumerate(arn_parts):
            self._parts[i] = arn_part

    def __repr__(self):
        return ":".join(self._parts)

    @property
    def parts(self):
        return self._parts

    def __eq__(self, value: Any):
        if not isinstance(value, _Arn):
            return False

        for i in range(0, 5):
            if self.parts[i] == "*":
                continue
            if value.parts[i] in ["${Partition}", "${Region}", "${Account}"]:
                continue
            if self.parts[i] != value.parts[i]:
                return False

        if self.parts[5] == "*":
            return True

        delimiter = None
        if ":" in value.parts[5]:
            delimiter = ":"
        elif "/" in value.parts[5]:
            delimiter = "/"

        if not delimiter:
            return True

        for x, y in zip(
            self.parts[5].split(delimiter), value.parts[5].split(delimiter)
        ):
            if x == y:
                continue
            if x == "*" or x.startswith("*") or y == "" or y == ".*":
                return True
            if x.startswith(y) and "*" in x:
                return True

            return False

        return True


class StatementResources(CfnLintKeyword):

    id = "I3510"
    shortdesc = "Validate statement resources match the actions"
    description = (
        "IAM policy statements have different constraints between actions and "
        "resources. This rule validates that resource ARNs or asterisks match "
        "the actions."
    )
    source_url = "https://docs.aws.amazon.com/AmazonCloudWatch/latest/logs/iam-access-control-overview-cwl.html"
    tags = ["resources", "iam"]

    def __init__(self):
        super().__init__(
            ["AWS::IAM::Policy/Properties/PolicyDocument/Statement"],
        )
        self.service_map = load_resource(AdditionalSpecs, "Policies.json")

    def validate(
        self, validator: Validator, keywords: Any, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:

        if not validator.is_type(instance, "object"):
            return

        using_fn_arns = False
        all_resources: set[str] = set()
        for key in ["Resource", "NotResource"]:
            for resources, _ in get_value_from_path(
                validator, instance, path=deque([key])
            ):
                resources = ensure_list(resources)

                for resource in resources:
                    if not isinstance(resource, str):
                        k, v = is_function(resource)
                        if k is None:
                            continue
                        if k == "Ref" and validator.is_type(v, "string"):
                            if v in validator.context.parameters:
                                return
                        using_fn_arns = True
                        continue
                    if resource == "*":
                        return
                    all_resources.add(resource)

        all_resource_arns = [_Arn(a) for a in all_resources]
        for actions, _ in get_value_from_path(
            validator, instance, path=deque(["Action"])
        ):
            actions = ensure_list(actions)

            for action in actions:

                if not validator.is_type(action, "string"):
                    continue
                if any(x in action for x in ["*", "?"]):
                    continue
                if ":" not in action:
                    continue
                service, permission = action.split(":", 1)
                service = service.lower()
                permission = permission.lower()

                if service not in self.service_map:
                    continue

                if permission not in self.service_map[service]["Actions"]:
                    continue

                resources = self.service_map[service]["Actions"][permission].get(
                    "Resources", None
                )
                if isinstance(resources, (list, str)):
                    if using_fn_arns:
                        continue
                    resources = ensure_list(resources)
                    for resource in resources:
                        # use lower as certain resources are described in actions
                        # with upper case letters and in resources as lower
                        # case issue #4036
                        arn_formats = self.service_map[service]["Resources"][
                            resource.lower()
                        ].get("ARNFormats")
                        for arn_format in arn_formats:
                            arn = _Arn(arn_format)
                        if arn in all_resource_arns:
                            break
                    else:
                        yield ValidationError(
                            (
                                f"action {action!r} requires "
                                f"a resource of {arn_formats!r}"
                            ),
                            path=deque(["Resource"]),
                            rule=self,
                        )
                else:
                    LOGGER.debug(f"action {action!r} requires a resource of '*'")
                    # yield ValidationError(
                    #    f"action {action!r} requires a resource of '*'",
                    #    path=deque(["Resource"]),
                    #    rule=self,
                    # )



"""
Copyright (c) 2013 Julian Berman

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

SPDX-License-Identifier: MIT
"""

# Code is taken from jsonschema package and adapted CloudFormation use
# https://github.com/python-jsonschema/jsonschema
from __future__ import annotations

import datetime
import numbers
from dataclasses import dataclass, field
from typing import Any, Callable

from cfnlint.jsonschema.exceptions import UndefinedTypeCheck


def is_array(checker: "TypeChecker", instance: Any) -> bool:
    return isinstance(instance, list)


def is_bool(checker: "TypeChecker", instance: Any) -> bool:
    return isinstance(instance, bool)


def is_integer(checker: "TypeChecker", instance: Any) -> bool:
    # bool inherits from int, so ensure bools aren't reported as ints
    if isinstance(instance, bool):
        return False
    return isinstance(instance, int)


def is_null(checker: "TypeChecker", instance: Any) -> bool:
    return instance is None


def is_number(checker: "TypeChecker", instance: Any) -> bool:
    # bool inherits from int, so ensure bools aren't reported as ints
    if isinstance(instance, bool):
        return False
    return isinstance(instance, numbers.Number)


def is_object(checker: "TypeChecker", instance: Any) -> bool:
    return isinstance(instance, dict)


def is_string(checker: "TypeChecker", instance: Any) -> bool:
    return isinstance(instance, str)


def is_date(checker: "TypeChecker", instance: Any) -> bool:
    return isinstance(instance, datetime.date)


@dataclass(frozen=True, repr=False)
class TypeChecker:
    """
    A :kw:`type` property checker.

    A `TypeChecker` performs type checking for a `Validator`, converting
    between the defined JSON Schema types and some associated Python types or
    objects.

    Modifying the behavior just mentioned by redefining which Python objects
    are considered to be of which JSON Schema types can be done using
    `TypeChecker.redefine` or `TypeChecker.redefine_many`, and types can be
    removed via `TypeChecker.remove`. Each of these return a new `TypeChecker`.

    Arguments:

        type_checkers:

            The initial mapping of types to their checking functions.
    """

    type_checkers: dict[str, Callable[[TypeChecker, Any], bool]] = field(
        init=True, default_factory=dict
    )

    def __repr__(self):
        types = ", ".join(repr(k) for k in sorted(self.type_checkers))
        return f"<{self.__class__.__name__} types={{{types}}}>"

    def is_type(self, instance, type: str) -> bool:
        """
        Check if the instance is of the appropriate type.

        Arguments:

            instance:

                The instance to check

            type:

                The name of the type that is expected.

        Raises:

            `cfnlint.jsonschema.exceptions.UndefinedTypeCheck`:

                if ``type`` is unknown to this object.
        """
        try:
            fn = self.type_checkers[type]
        except KeyError:
            raise UndefinedTypeCheck(type) from None

        return fn(self, instance)


cfn_type_checker = TypeChecker(
    {
        "array": is_array,
        "boolean": is_bool,
        "integer": lambda checker, instance: (
            is_integer(checker, instance)
            or isinstance(instance, float)
            and instance.is_integer()
        ),
        "object": is_object,
        "null": is_null,
        "number": is_number,
        "string": is_string,
        "date": is_date,
    },
)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.iam.ResourcePolicyResourceArn import (
    ResourcePolicyResourceArn,
)


@pytest.fixture(scope="module")
def rule():
    rule = ResourcePolicyResourceArn()
    yield rule


@pytest.mark.parametrize(
    "name,instance,path, expected",
    [
        (
            "Not a string",
            {},
            {"cfn_path": deque(["Resources", "AWS::S3::BucketPolicy", "Properties"])},
            [],
        ),
        (
            "Not long enough cfn path",
            "bad",
            {"cfn_path": deque(["Resources"])},
            [],
        ),
        (
            "Valid S3 ARN",
            "arn:aws:s3:::amzn-s3-demo-bucket",
            {"cfn_path": deque(["Resources", "AWS::S3::BucketPolicy", "Properties"])},
            [],
        ),
        (
            "Valid S3 ARN for objects",
            "arn:aws:s3:::amzn-s3-demo-bucket/*",
            {"cfn_path": deque(["Resources", "AWS::S3::BucketPolicy", "Properties"])},
            [],
        ),
        (
            "Invalid S3 resource ARN",
            "*",
            {"cfn_path": deque(["Resources", "AWS::S3::BucketPolicy", "Properties"])},
            [
                ValidationError(
                    (
                        "'*' does not match "
                        "'^arn:aws[A-Za-z\\\\-]*?:[^:]+:[^:]*(:(?:\\\\d{12}|\\\\*|aws)?:.+|)$'"
                    ),
                    rule=ResourcePolicyResourceArn(),
                )
            ],
        ),
        (
            "Invalid S3 resource ARN",
            "arn:aws*",
            {"cfn_path": deque(["Resources", "AWS::S3::BucketPolicy", "Properties"])},
            [
                ValidationError(
                    (
                        "'arn:aws*' does not match "
                        "'^arn:aws[A-Za-z\\\\-]*?:[^:]+:[^:]*(:(?:\\\\d{12}|\\\\*|aws)?:.+|)$'"
                    ),
                    rule=ResourcePolicyResourceArn(),
                )
            ],
        ),
        (
            "Valid SQS policy ARNs",
            "arn:aws:sqs:us-east-2:444455556666:queue1",
            {"cfn_path": deque(["Resources", "AWS::SQS::QueuePolicy", "Properties"])},
            [],
        ),
        (
            "Valid SQS policy ARNs",
            "*",
            {"cfn_path": deque(["Resources", "AWS::SQS::QueuePolicy", "Properties"])},
            [],
        ),
        (
            "Invalid SQS resource ARN",
            "arn:aws*",
            {"cfn_path": deque(["Resources", "AWS::SQS::QueuePolicy", "Properties"])},
            [
                ValidationError(
                    (
                        "'arn:aws*' does not match "
                        "'^(arn:aws[A-Za-z\\\\-]*?:[^:]+:[^:]*(:(?:\\\\d{12}|\\\\*|aws)?:.+|)|\\\\*)$'"
                    ),
                    rule=ResourcePolicyResourceArn(),
                )
            ],
        ),
    ],
    indirect=["path"],
)
def test_validate(name, instance, path, expected, rule, validator):
    errors = list(rule.validate(validator, {}, instance, {}))

    assert errors == expected, f"For test {name!r} got {errors!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.unit.rules import BaseRuleTestCase

from cfnlint.rules.custom.Operators import CreateCustomIsDefinedRule


class TestIsDefinedRule(BaseRuleTestCase):
    """Test template mapping configurations"""

    def setUp(self):
        """Setup"""
        super(TestIsDefinedRule, self).setUp()
        self.collection.register(
            CreateCustomIsDefinedRule(
                "E9001",
                "AWS::Lambda::Function",
                "Environment.Variables.NODE_ENV",
                "DEFINED",
                "NODE_ENV should be defined",
            )
        )

    success_templates = ["test/fixtures/templates/good/custom/is-defined.yaml"]

    def test_file_positive(self):
        """Test Positive"""
        self.helper_file_positive()

    def test_file_negative(self):
        """Test failure"""
        self.helper_file_negative(
            "test/fixtures/templates/good/custom/is-not-defined.yaml", 4
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from typing import Any

from cfnlint.jsonschema import ValidationResult, Validator
from cfnlint.rules.resources.iam.Policy import Policy


class ResourcePolicy(Policy):
    """Check IAM resource Policies"""

    id = "E3512"
    shortdesc = "Validate resource based IAM polices"
    description = (
        "IAM resources polices are embedded JSON in CloudFormation. "
        "This rule validates those embedded policies."
    )
    source_url = "https://docs.aws.amazon.com/IAM/latest/UserGuide/access_policies_identity-vs-resource.html"
    tags = ["resources", "iam"]

    def __init__(self):
        super().__init__(
            [
                "Resources/AWS::KMS::Key/Properties/KeyPolicy",
                "Resources/AWS::OpenSearchService::Domain/Properties/AccessPolicies",
                "Resources/AWS::S3::BucketPolicy/Properties/PolicyDocument",
                "Resources/AWS::SNS::TopicPolicy/Properties/PolicyDocument",
                "Resources/AWS::SQS::QueuePolicy/Properties/PolicyDocument",
            ],
            "resource",
            "policy_resource.json",
        )

    def validate(
        self,
        validator: Validator,
        policy_type: Any,
        policy: Any,
        schema: dict[str, Any],
    ) -> ValidationResult:
        for err in super().validate(validator, policy_type, policy, schema):
            if validator.context.path.cfn_path[1] == "AWS::S3::BucketPolicy":
                if err.validator == "uniqueKeys" and err.schema_path[-2] == "Statement":
                    continue
            yield err



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.ectwo.SecurityGroupProtocolsAndPortsExclusive import (
    SecurityGroupProtocolsAndPortsExclusive,
)


@pytest.fixture(scope="module")
def rule():
    rule = SecurityGroupProtocolsAndPortsExclusive()
    yield rule


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            {
                "IpProtocol": -1,
            },
            [],
        ),
        (
            {
                "IpProtocol": 9,
            },
            [],
        ),
        (
            {
                "IpProtocol": "TCP",
            },
            [],
        ),
        (
            [],  # wrong type
            [],
        ),
        (
            {
                "IpProtocol": -1,
                "FromPort": -1,
                "ToPort": -1,
            },
            [
                ValidationError(
                    (
                        "['FromPort', 'ToPort'] are ignored when using "
                        "'IpProtocol' value -1"
                    ),
                    rule=SecurityGroupProtocolsAndPortsExclusive(),
                    path=deque(["FromPort"]),
                    instance=-1,
                    schema=False,
                    validator=None,
                    validator_value=None,
                    schema_path=deque(["then", "properties", "FromPort"]),
                )
            ],
        ),
    ],
)
def test_backup_lifecycle(instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))

    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from dataclasses import dataclass, field, fields
from typing import TYPE_CHECKING, Any, Iterator, Sequence, Tuple

from cfnlint.helpers import FUNCTIONS, REGEX_DYN_REF, ToPy, ensure_list

if TYPE_CHECKING:
    from cfnlint.jsonschema.protocols import Validator

_all_types = ["array", "boolean", "integer", "number", "object", "string"]


@dataclass
class FunctionFilter:
    """
    A :kw:`type` property checker.

    A `FunctionChecker` looks for CloudFormation functions and resolves them
    as it can. There are scenarios where the value cannot be fully resolved
    and in that case nothing will happen

    Arguments:

        type_checkers:

            The initial mapping of types to their checking functions.
    """

    group_functions: Sequence[str] = field(
        init=False,
        default_factory=lambda: [
            "dependencies",
            "dependentExcluded",
            "dependentRequired",
            "else",
            "if",
            "maxItems",
            "maxProperties",
            "minItems",
            "minProperties",
            "required",
            "requiredOr",
            "requiredXor",
            "then",
            "uniqueItems",
        ],
    )

    add_cfn_lint_keyword: bool = field(
        init=True,
        default=True,
    )

    validate_dynamic_references: bool = field(
        init=True,
        default=True,
    )

    def _filter_schemas(self, schema, validator: Validator) -> Tuple[Any, Any]:
        """
        Filter the schemas to only include the ones that are required
        """

        # If we are in an if we want to keep any !Ref AWS::NoValue
        # Example: Typically we want to remove (!Ref AWS::NoValue)
        # to count minItems, maxItems, required properties but if we
        # are in an If we need to be more strict
        if validator.context.path.path:
            if validator.context.path.path[-1] in ["Fn::If"]:
                return schema, None

        standard_schema = {}
        group_schema = {}
        for key, value in schema.items():
            if key in self.group_functions:
                group_schema[key] = value
            else:
                standard_schema[key] = value
        if self.add_cfn_lint_keyword and "$ref" not in standard_schema:
            standard_schema["cfnLint"] = ensure_list(standard_schema.get("cfnLint", []))
            standard_schema["cfnLint"].append("/".join(validator.context.path.cfn_path))

        # some times CloudFormation dumps to standard nested "json".
        # it will do by using {"type": "object"} with no properties
        # Adding these items to the schema
        # will allow us to continue to check the nested elements
        # If we have the cfnLint keyword we assume that another check will
        # take care of this for us
        if "object" in ensure_list(standard_schema.get("type", [])) and all(
            p not in standard_schema
            for p in ["properties", "additionalProperties", "patternProperties"]
        ):
            standard_schema["patternProperties"] = {
                ".*": {"type": _all_types},
            }
        if (
            "array" in standard_schema.get("type", [])
            and "items" not in standard_schema
        ):
            standard_schema["items"] = {"type": _all_types}

        return standard_schema, group_schema

    def filter(
        self, validator: Any, instance: Any, schema: Any
    ) -> Iterator[tuple[Any, dict[str, Any], "Validator"]]:
        # if cfnContext is in the schema it is singular and
        # needs to be processed first
        if any(s in schema for s in ["cfnContext"]):
            yield instance, schema, validator
            return

        # Lets validate dynamic references when appropriate
        if validator.is_type(instance, "string") and self.validate_dynamic_references:
            if REGEX_DYN_REF.findall(instance):
                # if we are in a function we can't validate
                # dynamic references the same way
                if not any(
                    p in set(FUNCTIONS) - set(["Fn::If"])
                    for p in validator.context.path.path
                ):
                    yield (instance, {"dynamicReference": schema}, validator)
                    return
                return

        # if there are no functions then we don't need to worry
        # about ref AWS::NoValue or If conditions
        if not validator.context.functions:
            yield instance, schema, validator
            return

        # dependencies, required, minProperties, maxProperties
        # need to have filtered properties to validate
        # because of Ref: AWS::NoValue
        standard_schema, group_schema = self._filter_schemas(schema, validator)

        if group_schema:
            for (
                scenario_instance,
                scenario,
            ) in validator.context.conditions.evolve_from_instance(
                instance,
                context=validator.context,
            ):
                if scenario_instance is None:
                    continue
                scenario_validator = validator.evolve(
                    context=validator.context.evolve(conditions=scenario)
                )
                yield (scenario_instance, group_schema, scenario_validator)

        if validator.is_type(instance, "object"):
            if len(instance) == 1:
                for k, v in instance.items():
                    if k in validator.context.functions:
                        k_py = ToPy(k)
                        k_schema = {
                            k_py.py: standard_schema,
                        }
                        yield (instance, k_schema, validator)
                        return

        yield (instance, standard_schema, validator)

    def evolve(self, **kwargs) -> "FunctionFilter":
        """
        Create a new context merging together attributes
        """
        cls = self.__class__
        for f in fields(FunctionFilter):
            if f.init:
                kwargs.setdefault(f.name, getattr(self, f.name))

        return cls(**kwargs)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from cfnlint.rules.formats.FormatKeyword import FormatKeyword


class LogGroupName(FormatKeyword):
    id = "E1155"
    shortdesc = "Validate CloudWatch logs group name"
    description = "Check that a CloudWatch log group name matches a pattern"
    tags = []
    source_url = "https://github.com/aws-cloudformation/cfn-lint/blob/main/docs/format_keyword.md#AWS::Logs::LogGroup.Name"

    def __init__(self):
        super().__init__(
            format="AWS::EC2::Subnet.Id", pattern=r"^[\.\-_\/#A-Za-z0-9]{1,512}\Z"
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

import cfnlint.data.schemas.extensions.aws_ecs_service
from cfnlint.jsonschema import ValidationResult
from cfnlint.jsonschema.protocols import Validator
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails


class ServiceHealthCheckGracePeriodSeconds(CfnLintJsonSchema):
    id = "E3056"
    shortdesc = (
        "ECS service using HealthCheckGracePeriodSeconds "
        "must also have LoadBalancers specified"
    )
    description = (
        "When using a HealthCheckGracePeriodSeconds on an ECS "
        "service, the service must also have a LoadBalancers specified "
        "with at least one LoadBalancer in the array."
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-ecs-service.html#cfn-ecs-service-healthcheckgraceperiodseconds"
    tags = ["properties", "ecs", "service", "container"]

    def __init__(self) -> None:
        super().__init__(
            keywords=["Resources/AWS::ECS::Service/Properties"],
            schema_details=SchemaDetails(
                module=cfnlint.data.schemas.extensions.aws_ecs_service,
                filename="healthcheckgraceperiodseconds.json",
            ),
            all_matches=True,
        )

    def validate(
        self, validator: Validator, keywords: Any, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:

        cfn_validator = self.extend_validator(
            validator=validator.evolve(
                function_filter=validator.function_filter.evolve(
                    validate_dynamic_references=False,
                    add_cfn_lint_keyword=False,
                )
            ),
            schema=self._schema,
            context=validator.context.evolve(
                functions=["Fn::If", "Ref"],
                strict_types=True,
            ),
        )

        yield from self._iter_errors(cfn_validator, instance)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from cfnlint.rules.functions._BaseFn import BaseFn


class Cidr(BaseFn):
    """Check if Cidr values are correct"""

    id = "E1024"
    shortdesc = "Cidr validation of parameters"
    description = "Making sure the function CIDR is a list with valid values"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-cidr.html"
    tags = ["functions", "cidr"]

    def __init__(self) -> None:
        super().__init__(
            "Fn::Cidr",
            ("array",),
            None,
        )
        self.fn_cidr = self.validate



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import pytest

from cfnlint.context import create_context_for_template
from cfnlint.decode.cfn_yaml import loads
from cfnlint.rules import RuleMatch
from cfnlint.rules.aws_cli.UsingMerge import UsingMerge


@pytest.fixture(scope="module")
def rule():
    rule = UsingMerge()
    yield rule


@pytest.fixture(scope="module")
def context(cfn):
    return create_context_for_template(cfn)


@pytest.mark.parametrize(
    "name,template,expected",
    [
        (
            "A good template",
            loads(
                """
            One:
                A: 1
            """
            ),
            [],
        ),
        (
            "A merge template",
            loads(
                """
            One:
                &foo
                A: 1
            Two:
                <<: *foo
                A: 2
            """
            ),
            [
                RuleMatch(
                    path=["Two"],
                    message=(
                        "This code is using yaml marge capabilities "
                        "and can only be deployed using the "
                        "'package' cli command"
                    ),
                )
            ],
        ),
    ],
    indirect=["template"],
)
def test_validate(name, template, expected, rule, cfn):
    errs = list(rule.match(cfn))

    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

from cfnlint.helpers import REGIONS
from cfnlint.jsonschema import Validator
from cfnlint.rules.functions._BaseFn import BaseFn


class GetAz(BaseFn):
    """Check if GetAz values are correct"""

    id = "E1015"
    shortdesc = "GetAz validation of parameters"
    description = "Making sure the GetAz function is properly configured"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-getavailabilityzones.html"
    tags = ["functions", "getaz"]

    def __init__(self) -> None:
        super().__init__("Fn::GetAZs", ("array",), ("Ref",), resolved_rule="W1036")
        self.fn_getazs = self.validate

    def schema(self, validator: Validator, instance: Any) -> dict[str, Any]:
        return {
            "type": ["string"],
            "enum": [""] + REGIONS,
        }



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import unittest

from cfnlint.config import _merge_configs


class TestMergeConfigs(unittest.TestCase):
    """Test _merge_configs function"""

    def test_merge_lists(self):
        """Test merging lists"""
        cli_value = ["a", "b"]
        template_value = ["c", "d"]
        file_value = ["e", "f"]
        manual_value = ["g", "h"]

        result = _merge_configs(cli_value, template_value, file_value, manual_value)
        self.assertEqual(result, ["a", "b", "c", "d", "e", "f", "g", "h"])

    def test_merge_lists_with_none(self):
        """Test merging lists with None values"""
        cli_value = ["a", "b"]
        template_value = None
        file_value = ["e", "f"]
        manual_value = None

        result = _merge_configs(cli_value, template_value, file_value, manual_value)
        self.assertEqual(result, ["a", "b", "e", "f"])

    def test_merge_lists_with_non_lists(self):
        """Test merging lists with non-list values"""
        cli_value = ["a", "b"]
        template_value = "not a list"
        file_value = {"key": "value"}
        manual_value = 123

        result = _merge_configs(cli_value, template_value, file_value, manual_value)
        self.assertEqual(result, ["a", "b"])

    def test_merge_dicts(self):
        """Test merging dictionaries"""
        cli_value = {"a": 1, "b": 2}
        template_value = {"c": 3, "d": 4}
        file_value = {"e": 5, "f": 6}
        manual_value = {"g": 7, "h": 8}

        result = _merge_configs(cli_value, template_value, file_value, manual_value)
        self.assertEqual(
            result, {"a": 1, "b": 2, "c": 3, "d": 4, "e": 5, "f": 6, "g": 7, "h": 8}
        )

    def test_merge_dicts_with_none(self):
        """Test merging dictionaries with None values"""
        cli_value = {"a": 1, "b": 2}
        template_value = None
        file_value = {"e": 5, "f": 6}
        manual_value = None

        result = _merge_configs(cli_value, template_value, file_value, manual_value)
        self.assertEqual(result, {"a": 1, "b": 2, "e": 5, "f": 6})

    def test_merge_dicts_with_non_dicts(self):
        """Test merging dictionaries with non-dict values"""
        cli_value = {"a": 1, "b": 2}
        template_value = "not a dict"
        file_value = ["e", "f"]
        manual_value = 123

        result = _merge_configs(cli_value, template_value, file_value, manual_value)
        self.assertEqual(result, {"a": 1, "b": 2})

    def test_merge_dicts_with_overlapping_keys(self):
        """Test merging dictionaries with overlapping keys"""
        cli_value = {"a": 1, "b": 2, "common": "cli"}
        template_value = {"c": 3, "d": 4, "common": "template"}
        file_value = {"e": 5, "f": 6, "common": "file"}
        manual_value = {"g": 7, "h": 8, "common": "manual"}

        result = _merge_configs(cli_value, template_value, file_value, manual_value)
        # Last one in wins for overlapping keys
        self.assertEqual(
            result,
            {
                "a": 1,
                "b": 2,
                "c": 3,
                "d": 4,
                "e": 5,
                "f": 6,
                "g": 7,
                "h": 8,
                "common": "manual",
            },
        )

    def test_merge_non_list_non_dict(self):
        """Test merging when cli_value is neither list nor dict"""
        cli_value = "string"
        template_value = ["c", "d"]
        file_value = {"e": 5}
        manual_value = 123

        result = _merge_configs(cli_value, template_value, file_value, manual_value)
        self.assertIsNone(result)

    def test_merge_empty_values(self):
        """Test merging with empty values"""
        cli_value = []
        template_value = []
        file_value = []
        manual_value = []

        result = _merge_configs(cli_value, template_value, file_value, manual_value)
        self.assertEqual(result, [])

        cli_value = {}
        template_value = {}
        file_value = {}
        manual_value = {}

        result = _merge_configs(cli_value, template_value, file_value, manual_value)
        self.assertEqual(result, {})



"""
Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import logging
from typing import Any, Dict

from sympy import And, Implies, Symbol
from sympy.logic.boolalg import BooleanFunction, BooleanTrue

from cfnlint.conditions._condition import (
    ConditionAnd,
    ConditionList,
    ConditionNamed,
    ConditionNot,
    ConditionOr,
)
from cfnlint.conditions._equals import Equal
from cfnlint.helpers import FUNCTION_CONDITIONS

LOGGER = logging.getLogger(__name__)

# we leave the type hinting here
_RULE = Dict[str, Any]


class _Assertion:
    def __init__(self, condition: Any, all_conditions: dict[str, dict]) -> None:
        self._fn_equals: Equal | None = None
        self._condition: ConditionList | ConditionNamed | None = None

        if len(condition) == 1:
            for k, v in condition.items():
                if k in FUNCTION_CONDITIONS:
                    if not isinstance(v, list):
                        raise ValueError(f"{k} value should be an array")
                    if k == "Fn::Equals":
                        self._fn_equals = Equal(v)
                    elif k == "Fn::And":
                        self._condition = ConditionAnd(v, all_conditions)
                    elif k == "Fn::Or":
                        self._condition = ConditionOr(v, all_conditions)
                    elif k == "Fn::Not":
                        self._condition = ConditionNot(v, all_conditions)
                elif k == "Condition":
                    if not isinstance(v, str):
                        raise ValueError(f"Condition value {v!r} must be a string")
                    self._condition = ConditionNamed(v, all_conditions)
                else:
                    raise ValueError(f"Unknown key ({k}) in condition")
        else:
            raise ValueError("Condition value must be an object of length 1")

    def build_cnf(self, params: dict[str, Symbol]) -> BooleanFunction | Symbol | None:
        if self._fn_equals:
            try:
                return self._fn_equals.build_cnf(params)
            except Exception as e:
                LOGGER.debug(f"Error building condition: {e}")

        if self._condition:
            try:
                return self._condition.build_cnf(params)
            except Exception as e:
                LOGGER.debug(f"Error building condition: {e}")

        return BooleanTrue()

    @property
    def equals(self) -> list[Equal]:
        if self._fn_equals:
            return [self._fn_equals]
        if self._condition:
            return self._condition.equals
        return []


class _Assertions:
    def __init__(self, assertions: list[dict], all_conditions: dict[str, dict]) -> None:
        self._assertions: list[_Assertion] = []
        for assertion in assertions:
            assert_ = assertion.get("Assert", {})
            self._assertions.append(_Assertion(assert_, all_conditions))

    def build_cnf(self, params: dict[str, Symbol]) -> BooleanFunction | Symbol | None:

        assertions = []
        for assertion in self._assertions:
            assertions.append(assertion.build_cnf(params))

        try:
            return And(*assertions)
        except Exception as e:
            LOGGER.debug(f"Error building conditions: {e}")
            return BooleanTrue()

    @property
    def equals(self) -> list[Equal]:

        results = []
        for assertion in self._assertions:
            results.extend(assertion.equals)
        return results


class Rule:

    def __init__(self, rule: _RULE, all_conditions: dict[str, dict]) -> None:
        self._condition: _Assertion | None = None
        self._assertions: _Assertions | None = None
        self._init_rule(rule, all_conditions)

    def _init_rule(
        self,
        rule: _RULE,
        all_conditions: dict[str, dict],
    ) -> None:
        condition = rule.get("RuleCondition")
        if condition:
            self._condition = _Assertion(condition, all_conditions)

        assertions = rule.get("Assertions")
        if not assertions:
            raise ValueError("Rule must have Assertions")
        self._assertions = _Assertions(assertions, all_conditions)

    @property
    def equals(self) -> list[Equal]:
        result = []
        if self._condition:
            result.extend(self._condition.equals)
        if self._assertions:
            result.extend(self._assertions.equals)
        return result

    def build_cnf(self, params: dict[str, Symbol]) -> BooleanFunction | Symbol | None:

        if self._assertions:
            if self._condition:
                return Implies(
                    self._condition.build_cnf(params),
                    self._assertions.build_cnf(params),
                )
            return self._assertions.build_cnf(params)
        return None



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.functions.DynamicReferenceSecretsManagerArn import (
    DynamicReferenceSecretsManagerArn,
)


@pytest.fixture(scope="module")
def rule():
    rule = DynamicReferenceSecretsManagerArn()
    yield rule


@pytest.mark.parametrize(
    "name,instance,path,expected",
    [
        (
            "Valid secrets manager",
            "{{resolve:secretsmanager:Parameter}}",
            {
                "path": deque(["Resources", "MyResource", "Properties", "Password"]),
            },
            [],
        ),
        (
            "Valid sub not to secrets manaager",
            "secretsmanager",
            {
                "path": deque(
                    ["Resources", "MyResource", "Properties", "Password", "Fn::Sub"]
                ),
            },
            [],
        ),
        (
            "Valid secrets manager outside of Resources",
            "{{resolve:secretsmanager:Parameter}}",
            {
                "path": deque(["Parameters", "Type", "Default"]),
            },
            [],
        ),
        (
            "Valid secrets manager outside of Resources",
            "{{resolve:secretsmanager:Parameter}}",
            {
                "path": deque(["Metadata", "Value"]),
            },
            [],
        ),
        (
            "Invalid secrets manager when ARN was expected",
            "{{resolve:secretsmanager:secret}}",
            {
                "path": deque(["Resources", "MyResource", "Properties", "SecretArn"]),
            },
            [
                ValidationError(
                    (
                        "Dynamic reference '{{resolve:secretsmanager:secret}}'"
                        " to secrets manager when the field 'SecretArn' expects "
                        "the ARN to the secret and not the secret"
                    ),
                    rule=DynamicReferenceSecretsManagerArn(),
                )
            ],
        ),
        (
            "Invalid secrets manager when ARN was expected",
            '\\"{{resolve:secretsmanager:${MyParameter}}}\\"',
            {
                "path": deque(
                    ["Resources", "MyResource", "Properties", "SecretArn", "Fn::Sub"]
                ),
            },
            [
                ValidationError(
                    (
                        "Dynamic reference "
                        "'\\\\\"{{resolve:secretsmanager:${MyParameter}}}\\\\\"'"
                        " to secrets manager when the field 'SecretArn' "
                        "expects the ARN to the secret and not the secret"
                    ),
                    rule=DynamicReferenceSecretsManagerArn(),
                )
            ],
        ),
    ],
    indirect=["path"],
)
def test_validate(name, instance, path, expected, validator, rule):
    errs = list(rule.validate(validator, {}, instance, {}))

    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from typing import Any

import regex as re

from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword


class AllowedPattern(CfnLintKeyword):

    id = "I2003"
    shortdesc = "Validate AllowedPattern is a valid regexs"
    description = (
        "Validate the pattern defined in a AllowedPattern. "
        "This is informational as the service side regex library is "
        "different than the Python one"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/cloudformation-limits.html"
    tags = ["parameters", "allowed pattern"]

    def __init__(self) -> None:
        super().__init__(
            keywords=[
                "Parameters/*/AllowedPattern",
            ]
        )

    def validate(
        self, validator: Validator, keywords: Any, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:

        if not validator.is_type(instance, "string"):
            return

        try:
            re.compile(instance)
        except Exception as e:
            yield ValidationError(
                f"{instance!r} could not be compiled ({str(e)})", rule=self
            )



"""
Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations


class ResourceNotFoundError(Exception):
    def __init__(self, resource_type: str, region: str):
        super().__init__(f"Resource type '{resource_type}' is not found in '{region}'")


class SchemaNotFoundError(Exception):
    def __init__(self, path: str):
        super().__init__(f"Schema '{path}' is not found")



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import pathlib

from cfnlint.helpers import load_plugins


def test_cfn_lint_rules_have_validate_function():
    root_dir = pathlib.Path(__file__).parent.parent.parent.parent / "src/cfnlint/rules"
    rules = load_plugins(
        str(root_dir),
        "CfnLintKeyword",
        "cfnlint.rules.jsonschema.CfnLintKeyword",
    )
    rules.extend(
        load_plugins(
            str(root_dir),
            "CfnLintJsonSchema",
            "cfnlint.rules.jsonschema.CfnLintJsonSchema",
        )
    )
    rules.extend(
        load_plugins(
            str(root_dir),
            "CfnLintJsonSchemaRegional",
            "cfnlint.rules.jsonschema.CfnLintJsonSchemaRegional",
        )
    )

    parent_rules = [
        "E1010",
        "E1101",
        "E1020",
    ]

    for rule in rules:
        assert hasattr(rule, "validate")
        assert callable(rule.validate)
        assert all(
            r in parent_rules for r in rule._parent_rules
        ), f"{rule._parent_rules!r} not in {parent_rules!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.parameters.AllowedPattern import AllowedPattern


@pytest.fixture(scope="module")
def rule():
    rule = AllowedPattern()
    yield rule


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            {},
            [],
        ),
        (
            "^foo$",
            [],
        ),
        (
            "^abc|def)$",
            [
                ValidationError(
                    (
                        "'^abc|def)$' could not be "
                        "compiled (unbalanced parenthesis "
                        "at position 8)"
                    ),
                    rule=AllowedPattern(),
                )
            ],
        ),
    ],
)
def test_backup_lifecycle(instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))

    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import logging
from json.decoder import JSONDecodeError
from typing import Any, Callable, Dict, List, Tuple, Union

from yaml import YAMLError
from yaml.parser import ParserError
from yaml.scanner import ScannerError

from cfnlint.decode import cfn_json, cfn_yaml
from cfnlint.match import Match

LOGGER = logging.getLogger(__name__)

Matches = List[Match]
Decode = Tuple[Union[Dict[str, Any], None], Matches]


def decode_str(s: str) -> Decode:
    """Decode the string s into an object."""
    return _decode(cfn_yaml.loads, cfn_json.loads, s, None)


def decode(filename: str | None) -> Decode:
    """Decode filename into an object."""
    return _decode(cfn_yaml.load, cfn_json.load, filename, filename)


def _decode(
    yaml_f: Callable, json_f: Callable, payload: str | None, filename: str | None
) -> Decode:
    """Decode payload using yaml_f and json_f, using filename for log output."""
    template = None
    matches = []
    try:
        template = yaml_f(payload)
    except IOError as e:
        if e.errno == 2:
            LOGGER.error("Template file not found: %s", filename)
            matches.append(
                create_match_file_error(
                    filename, f"Template file not found: {filename}"
                )
            )
        elif e.errno == 21:
            LOGGER.error("Template references a directory, not a file: %s", filename)
            matches.append(
                create_match_file_error(
                    filename, f"Template references a directory, not a file: {filename}"
                )
            )
        elif e.errno == 13:
            LOGGER.error("Permission denied when accessing template file: %s", filename)
            matches.append(
                create_match_file_error(
                    filename,
                    f"Permission denied when accessing template file: {filename}",
                )
            )

        if matches:
            return (None, matches)
    except UnicodeDecodeError:
        LOGGER.error("Cannot read file contents: %s", filename)
        matches.append(
            create_match_file_error(filename, "Cannot read file contents: {filename}")
        )
    except cfn_yaml.CfnParseError as err:
        matches = err.matches
    except ParserError as err:
        matches = [create_match_yaml_parser_error(err, filename)]
    except ScannerError as err:
        if err.problem and (
            err.problem
            in [
                "found character '\\t' that cannot start any token",
                "found unknown escape character",
            ]
            or err.problem.startswith("found unknown escape character")
        ):
            try:
                template = json_f(payload)
            except cfn_json.JSONDecodeError as json_errs:
                for json_err in json_errs.matches:
                    matches.append(
                        Match(
                            message=json_err.message,
                            filename=filename,
                            linenumber=json_err.linenumber,
                            columnnumber=json_err.columnnumber,
                            linenumberend=json_err.linenumberend,
                            columnnumberend=json_err.columnnumberend,
                            rule=json_err.rule,
                            parent_id=json_err.parent_id,
                        )
                    )
            except JSONDecodeError as json_err:
                if hasattr(json_err, "msg"):
                    if json_err.msg in [
                        "No JSON object could be decoded",
                        "Expecting value",
                    ]:  # pylint: disable=no-member
                        matches = [create_match_yaml_parser_error(err, filename)]
                    else:
                        matches = [create_match_json_parser_error(json_err, filename)]
            except Exception as json_err:  # pylint: disable=W0703
                LOGGER.error("Template %s is malformed: %s", filename, err.problem)
                LOGGER.error(
                    "Tried to parse %s as JSON but got error: %s",
                    filename,
                    str(json_err),
                )
                return (
                    None,
                    [
                        create_match_file_error(
                            filename,
                            (
                                f"Tried to parse {filename} as JSON but got error:"
                                f" {str(json_err)}"
                            ),
                        )
                    ],
                )
        else:
            matches = [create_match_yaml_parser_error(err, filename)]
    except (YAMLError, Exception) as err:
        matches = [create_match_file_error(filename, str(err))]

    if not isinstance(template, dict) and not matches:
        # pylint: disable=import-outside-toplevel
        from cfnlint.rules.errors import ParseError

        # Template isn't a dict which means nearly nothing will work
        matches = [
            Match.create(
                filename=filename or "",
                rule=ParseError(),
                message="Template needs to be an object.",
            )
        ]
    return (template, matches)


def create_match_yaml_parser_error(parser_error, filename):
    """Create a Match for a parser error"""
    # pylint: disable=import-outside-toplevel
    from cfnlint.rules.errors import ParseError

    lineno = parser_error.problem_mark.line + 1
    colno = parser_error.problem_mark.column + 1
    msg = parser_error.problem
    return Match.create(
        message=msg,
        rule=ParseError(),
        filename=filename,
        linenumber=lineno,
        columnnumber=colno,
    )


def create_match_file_error(filename, msg):
    """Create a Match for a parser error"""
    # pylint: disable=import-outside-toplevel
    from cfnlint.rules.errors import ParseError

    return Match.create(
        filename=filename,
        rule=ParseError(),
        message=msg,
    )


def create_match_json_parser_error(parser_error, filename):
    """Create a Match for a parser error"""
    # pylint: disable=import-outside-toplevel
    from cfnlint.rules.errors import ParseError

    lineno = parser_error.lineno
    colno = parser_error.colno
    msg = parser_error.msg
    return Match.create(
        filename=filename,
        rule=ParseError(),
        message=msg,
        linenumber=lineno,
        columnnumber=colno,
    )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.context import Context, Path
from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.lmbd.SnapStart import SnapStart


@pytest.fixture(scope="module")
def rule():
    rule = SnapStart()
    yield rule


@pytest.fixture
def template():
    return {
        "Resources": {
            "GoodSnapStart": {
                "Type": "AWS::Lambda::Function",
                "Properties": {
                    "Code": {"S3Bucket": "XXXXXX", "S3Key": "key"},
                    "Handler": "handler",
                    "Role": "role",
                    "Runtime": "runtime",
                    "SnapStart": {"ApplyOn": "PublishedVersions"},
                },
            },
            "GoodSnapStartVersion": {
                "Type": "AWS::Lambda::Version",
                "Properties": {"FunctionName": {"Ref": "GoodSnapStart"}},
            },
            "BadSnapStart": {
                "Type": "AWS::Lambda::Function",
                "Properties": {
                    "Code": {"S3Bucket": "XXXXXX", "S3Key": "key"},
                    "Handler": "handler",
                    "Role": "role",
                    "Runtime": "runtime",
                    "SnapStart": {"ApplyOn": "PublishedVersions"},
                },
            },
        }
    }


@pytest.mark.parametrize(
    "name,instance,path,expected",
    [
        (
            "None type should result in no errors",
            "None",  # Not
            deque(["Resources", "BadSnapStart", "Properties", "SnapStart", "ApplyOn"]),
            [],
        ),
        (
            "Wrong type should result in no errors",
            [],  # wrong type
            deque(["Resources", "BadSnapStart", "Properties", "SnapStart", "ApplyOn"]),
            [],
        ),
        (
            "Correctly associated version to lambda function",
            "PublishedVersions",
            deque(["Resources", "GoodSnapStart", "Properties", "SnapStart", "ApplyOn"]),
            [],
        ),
        (
            "Lambda function doesn't have version attached",
            "PublishedVersions",
            deque(["Resources", "BadSnapStart", "Properties", "SnapStart", "ApplyOn"]),
            [
                ValidationError(
                    (
                        "'SnapStart' is enabled but an 'AWS::Lambda::Version' "
                        "resource is not attached"
                    ),
                )
            ],
        ),
    ],
)
def test_validate(name, instance, path, expected, rule, validator):
    validator = validator.evolve(
        context=Context(
            path=Path(
                path=path,
                cfn_path=deque(
                    [
                        "Resources",
                        "AWS::Lambda::Function",
                        "Properties",
                        "SnapStart",
                        "ApplyOn",
                    ]
                ),
            )
        )
    )
    errs = list(rule.validate(validator, "", instance, {}))

    assert errs == expected, f"{name!r}: expected {expected!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules import CloudFormationLintRule
from cfnlint.rules.functions._BaseFn import BaseFn


class _ChildRule(CloudFormationLintRule):
    id = "XXXXX"


@pytest.fixture(scope="module")
def rule(request):
    rule = BaseFn(resolved_rule="XXXXX")
    rule._schema = request.param.get("schema")

    rule.child_rules["XXXXX"] = _ChildRule()
    yield rule


@pytest.mark.parametrize(
    "name,rule,instance,expected",
    [
        (
            "Dynamic references are ignored",
            {
                "schema": {
                    "enum": ["Foo"],
                },
                "patches": [],
            },
            {"Fn::Sub": "{{resolve:ssm:${AWS::AccountId}/${AWS::Region}/ac}}"},
            [],
        ),
        (
            "Everything is fine",
            {
                "schema": {"enum": ["Foo"]},
                "patches": [],
            },
            {"Fn::Sub": "Foo"},
            [],
        ),
        (
            "Resolved Fn::Sub has no strict type validation",
            {
                "schema": {"type": ["integer"]},
                "patches": [],
            },
            {"Fn::Sub": "2"},
            [],
        ),
        (
            "Standard error",
            {
                "schema": {"enum": ["Foo"]},
                "patches": [],
            },
            {"Fn::Sub": "Bar"},
            [
                ValidationError(
                    message=(
                        "{'Fn::Sub': 'Bar'} is not one of "
                        "['Foo'] when '' is resolved"
                    ),
                    path=deque(["Fn::Sub"]),
                    validator="",
                    schema_path=deque(["enum"]),
                    rule=_ChildRule(),
                )
            ],
        ),
        (
            "Errors with context error",
            {
                "schema": {"anyOf": [{"enum": ["Foo"]}]},
                "patches": [],
            },
            {"Fn::Sub": "Bar"},
            [
                ValidationError(
                    message=(
                        "{'Fn::Sub': 'Bar'} is not valid "
                        "under any of the given schemas "
                        "when '' is resolved"
                    ),
                    path=deque(["Fn::Sub"]),
                    validator="",
                    schema_path=deque(["anyOf"]),
                    rule=_ChildRule(),
                    context=[
                        ValidationError(
                            message=(
                                "{'Fn::Sub': 'Bar'} is not one of "
                                "['Foo'] when '' is resolved"
                            ),
                            path=deque([]),
                            validator="enum",
                            schema_path=deque([0, "enum"]),
                            rule=_ChildRule(),
                        )
                    ],
                )
            ],
        ),
    ],
    indirect=["rule"],
)
def test_resolve(name, instance, expected, validator, rule):
    errs = list(rule.resolve(validator, rule.schema(validator, instance), instance, {}))
    assert errs == expected, f"{name!r} failed and got errors {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

from cfnlint.rules import CloudFormationLintRule, RuleMatch
from cfnlint.template import Template


class UsingMerge(CloudFormationLintRule):
    id = "W1100"
    shortdesc = "Validate if the template is using YAML merge"
    description = (
        "The CloudFormation service does not support YAML anchors, "
        "aliases, or merging. This rule validates if the "
        "merge capability is being used"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/template-formats.html"
    tags = ["yaml"]

    def _nested_obj(self, obj: Any, path: list[str | int]) -> list[RuleMatch]:
        matches = []
        if isinstance(obj, dict):
            if hasattr(obj, "using_merge") and obj.using_merge:
                matches.append(
                    RuleMatch(
                        path=path,
                        message=(
                            "This code is using yaml marge capabilities "
                            "and can only be deployed using the "
                            "'package' cli command"
                        ),
                    )
                )
            for k, v in obj.items():
                matches.extend(self._nested_obj(v, path + [k]))
        elif isinstance(obj, list):
            for i, e in enumerate(obj):
                matches.extend(self._nested_obj(e, path + [i]))
        return matches

    def match(self, cfn: Template):

        return self._nested_obj(cfn.template, [])



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from io import StringIO
from unittest.mock import patch

import pytest

from cfnlint.config import ConfigMixIn
from cfnlint.runner import PROVIDER_SCHEMA_MANAGER, Runner
from cfnlint.schema import Schema, reset


def patch_registry(path):
    PROVIDER_SCHEMA_MANAGER._registry_schemas = {
        "Foo::Bar::MODULE": Schema(
            {
                "typeName": "Foo::Bar::MODULE",
                "properties": {
                    "Foo": {
                        "type": "string",
                    },
                    "Bar": {
                        "type": "string",
                    },
                },
                "primaryIdentifiers": [],
            }
        )
    }


def patch_schema(path, regions):
    PROVIDER_SCHEMA_MANAGER._registry_schemas["Foo::Bar::MODULE"].schema[
        "primaryIdentifiers"
    ] = ["/properties/Foo"]


@pytest.mark.parametrize(
    "name,registry_path,patch_path,expected",
    [
        (
            "Test no changes to the schema",
            None,
            None,
            False,
        ),
        (
            "Test patching registry resources",
            "path/to/registry/schemas",
            "patch.json",
            False,
        ),
        (
            "Test registry resource with no patching",
            "path/to/registry/schemas",
            None,
            False,
        ),
    ],
)
def test_init_schemas(name, registry_path, patch_path, expected):
    params = []
    if registry_path:
        params.extend(["--registry-schemas", "path/to/registry/schemas"])
    if patch_path:
        params.extend(["--override-spec", "patch.json"])
    config = ConfigMixIn(params)

    with patch.object(
        PROVIDER_SCHEMA_MANAGER, "load_registry_schemas", new=patch_registry
    ):
        with patch("cfnlint.runner.patch", new=patch_schema):
            Runner(config)

            if registry_path:
                assert "Foo::Bar::MODULE" in PROVIDER_SCHEMA_MANAGER.get_resource_types(
                    "us-east-1"
                )
                schema = PROVIDER_SCHEMA_MANAGER.get_resource_schema(
                    "us-east-1", "Foo::Bar::MODULE"
                )
                if patch_path:
                    assert schema.schema["primaryIdentifiers"] == ["/properties/Foo"]
                else:
                    assert schema.schema["primaryIdentifiers"] == []
            else:
                assert (
                    "Foo::Bar::MODULE"
                    not in PROVIDER_SCHEMA_MANAGER.get_resource_types("us-east-1")
                )

    PROVIDER_SCHEMA_MANAGER._registry_schemas = {}
    reset()


def test_no_templates():
    params = ["--template", "does-not-exist.yaml"]

    config = ConfigMixIn(params)
    with patch("sys.exit") as exit:
        with patch("sys.stdout", new=StringIO()) as out:
            exit.assert_not_called()
            Runner(config)
            assert out.getvalue().strip() == (
                "E0003 does-not-exist.yaml could not "
                "be processed by glob.glob\nNone:1:1"
            )
            exit.assert_called_once_with(2)



"""
Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import json
import logging
import sys
from dataclasses import dataclass
from typing import Dict, Sequence

from cfnlint.schema.manager import PROVIDER_SCHEMA_MANAGER
from cfnlint.schema.other_schema_manager import OTHER_SCHEMA_MANAGER

LOGGER = logging.getLogger(__name__)


@dataclass
class SchemaPatch:
    included_resource_types: list[str]
    excluded_resource_types: list[str]
    patches: dict[str, list[Dict]]

    @staticmethod
    def from_dict(value: Dict):
        return SchemaPatch(
            included_resource_types=value.get("IncludeResourceTypes", []),
            excluded_resource_types=value.get("ExcludeResourceTypes", []),
            patches=value.get("Patches", {}),
        )


def patch(filename: str, regions: Sequence[str]):
    try:
        with open(filename, encoding="utf-8") as fp:
            custom_spec_data = json.load(fp)
            schema_patch = SchemaPatch.from_dict(custom_spec_data)
            for region in regions:
                PROVIDER_SCHEMA_MANAGER.patch(schema_patch, region)

            OTHER_SCHEMA_MANAGER.patch(schema_patch, region)
    except IOError as e:
        if e.errno == 2:
            LOGGER.error("Override spec file not found: %s", filename)
            sys.exit(1)
        elif e.errno == 21:
            LOGGER.error(
                "Override spec file references a directory, not a file: %s",
                filename,
            )
            sys.exit(1)
        elif e.errno == 13:
            LOGGER.error(
                "Permission denied when accessing override spec file: %s", filename
            )
            sys.exit(1)
    except ValueError as err:
        LOGGER.error("Override spec file %s is malformed: %s", filename, err)
        sys.exit(1)


def reset():
    PROVIDER_SCHEMA_MANAGER.reset()
    OTHER_SCHEMA_MANAGER.reset()



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.unit.rules import BaseRuleTestCase

from cfnlint.jsonschema import CfnTemplateValidator

# ruff: noqa: E501
from cfnlint.rules.resources.properties.Type import Type


class TestType(BaseRuleTestCase):
    """Test Primitive Value Types for Json Schema non strict"""

    def setUp(self):
        """Setup"""
        self.rule = Type()
        self.rule.config["strict"] = False
        self.validator = CfnTemplateValidator()

    def check_args(self, types, instance, extra_args):
        errs = list(self.rule.type(self.validator, types, instance, {}))
        self.assertEqual(len(errs), 1)
        self.assertDictEqual(
            errs[0].extra_args,
            extra_args,
        )

    def test_validation_non_strict(self):
        self.rule.config["strict"] = False
        # sub is a string boolean
        self.check_args(
            types="integer",
            instance="A",
            extra_args={
                "actual_type": "string",
                "expected_type": "integer",
            },
        )
        self.check_args(
            types=["number", "integer"],
            instance="A",
            extra_args={
                "actual_type": "string",
                "expected_type": "number",
            },
        )

    def test_validation_strict(self):
        self.rule.config["strict"] = True
        # sub is a string boolean
        self.check_args(
            types="integer",
            instance="1",
            extra_args={
                "actual_type": "string",
                "expected_type": "integer",
            },
        )
        self.check_args(
            types=["string", "boolean"],
            instance=1,
            extra_args={
                "actual_type": "integer",
                "expected_type": "string",
            },
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import jsonpatch
import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.lmbd.EventSourceMappingToSqsTimeout import (
    EventSourceMappingToSqsTimeout,
)


@pytest.fixture(scope="module")
def rule():
    rule = EventSourceMappingToSqsTimeout()
    yield rule


_template = {
    "Parameters": {
        "BatchSize": {"Type": "String"},
    },
    "Resources": {
        "MyFifoQueue": {
            "Type": "AWS::SQS::Queue",
            "Properties": {
                "VisibilityTimeout": 300,
                "MessageRetentionPeriod": 7200,
            },
        },
        "SQSBatch": {
            "Type": "AWS::Lambda::EventSourceMapping",
            "Properties": {
                "BatchSize": {"Ref": "BatchSize"},
                "Enabled": True,
                "EventSourceArn": {"Fn::GetAtt": "MyFifoQueue.Arn"},
                "FunctionName": {"Ref": "Lambda"},
            },
        },
        "Lambda": {
            "Type": "AWS::Lambda::Function",
            "Properties": {"Role": {"Fn::GetAtt": "Role.Arn"}},
        },
        "CustomResource": {
            "Type": "AWS::CloudFormation::CustomResource",
            "Properties": {"Key": {"Fn::GetAtt": "Lambda.Arn"}},
        },
    },
    "Outputs": {
        "LambdaArn": {"Value": {"Fn::GetAtt": "Lambda.Arn"}},
        "SourceMapping": {"Value": {"Ref": "SQSBatch"}},
    },
}


@pytest.mark.parametrize(
    "instance,template,path,expected",
    [
        (
            "100",
            _template,
            {"path": deque(["Resources", "Lambda", "Properties", "Timeout"])},
            [],
        ),
        (
            "a",
            _template,
            {"path": deque(["Resources", "Lambda", "Properties", "Timeout"])},
            [],
        ),
        (
            {"Ref": "AWS::Region"},
            _template,
            {"path": deque(["Resources", "Lambda", "Properties", "Timeout"])},
            [],
        ),
        (
            "100",
            jsonpatch.apply_patch(
                _template,
                [
                    {
                        "op": "add",
                        "path": (
                            "/Resources/MyFifoQueue/" "Properties/VisibilityTimeout"
                        ),
                        "value": "300",
                    },
                ],
            ),
            {"path": deque(["Resources", "Lambda", "Properties", "Timeout"])},
            [],
        ),
        (
            "600",
            jsonpatch.apply_patch(
                _template,
                [
                    {
                        "op": "add",
                        "path": (
                            "/Resources/MyFifoQueue/" "Properties/VisibilityTimeout"
                        ),
                        "value": {"Ref": "AWS::Region"},
                    },
                ],
            ),
            {"path": deque(["Resources", "Lambda", "Properties", "Timeout"])},
            [],
        ),
        (
            "600",
            jsonpatch.apply_patch(
                _template,
                [
                    {
                        "op": "add",
                        "path": (
                            "/Resources/MyFifoQueue/" "Properties/VisibilityTimeout"
                        ),
                        "value": "a",
                    },
                ],
            ),
            {"path": deque(["Resources", "Lambda", "Properties", "Timeout"])},
            [],
        ),
        (
            "600",
            _template,
            {"path": deque(["Resources"])},
            [],
        ),
        (
            "600",
            _template,
            {"path": deque(["Resources", "Lambda", "Properties", "Timeout"])},
            [
                ValidationError(
                    (
                        "Queue visibility timeout (300) is less "
                        "than Function timeout (600) seconds"
                    ),
                    rule=EventSourceMappingToSqsTimeout(),
                )
            ],
        ),
    ],
    indirect=["template", "path"],
)
def test_lambda_runtime(instance, template, path, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))
    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.unit.rules import BaseRuleTestCase

from cfnlint.rules.resources.dynamodb.AttributeMismatch import (
    AttributeMismatch,  # pylint: disable=E0401
)


class TestAttributeMismatch(BaseRuleTestCase):
    """Test AttributeMisMatch"""

    def setUp(self):
        """Setup"""
        super(TestAttributeMismatch, self).setUp()
        self.collection.register(AttributeMismatch())
        self.success_templates = [
            "test/fixtures/templates/good/resources/dynamodb/attributes.yaml"
        ]

    def test_file_positive(self):
        """Test Positive"""
        self.helper_file_positive()

    def test_file_negative_unused_1(self):
        """Test failure"""
        self.helper_file_negative(
            "test/fixtures/templates/bad/resources/dynamodb/unused_attribute_definition_1.yaml",
            1,
        )

    def test_file_negative_unused_2(self):
        """Test failure"""
        self.helper_file_negative(
            "test/fixtures/templates/bad/resources/dynamodb/unused_attribute_definition_2.yaml",
            1,
        )

    def test_file_negative_undefined(self):
        """Test failure"""
        self.helper_file_negative(
            "test/fixtures/templates/bad/resources/dynamodb/undefined_attribute_definition.yaml",
            1,
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import argparse
import copy
import glob
import json
import logging
import os
import sys
from pathlib import Path
from typing import Any, Dict, Sequence, TypedDict

from typing_extensions import Unpack

import cfnlint.decode.cfn_yaml
from cfnlint.helpers import REGIONS, format_json_string
from cfnlint.jsonschema import StandardValidator
from cfnlint.version import __version__

# pylint: disable=too-many-public-methods
LOGGER = logging.getLogger("cfnlint")

_DEFAULT_RULESDIR = os.path.join(os.path.dirname(__file__), "rules")


def configure_logging(debug_logging, info_logging):
    ch = logging.StreamHandler()

    if debug_logging:
        LOGGER.setLevel(logging.DEBUG)
    elif info_logging:
        LOGGER.setLevel(logging.INFO)
    else:
        LOGGER.setLevel(logging.WARNING)
    log_formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    ch.setFormatter(log_formatter)

    # make sure all other log handlers are removed before adding it back
    for handler in LOGGER.handlers:
        LOGGER.removeHandler(handler)
    LOGGER.addHandler(ch)


class ConfigFileArgs:
    """
    Config File arguments.
    Parses .cfnlintrc OR .cfnlintrc.yaml OR .cfnlintrc.yml
    in the Home and Project folder.
    """

    file_args: Dict = {}
    _user_config_file = None
    _project_config_file = None
    _custom_config_file = None

    def __init__(self, schema=None, config_file=None):
        # self.file_args = self.get_config_file_defaults()
        self.file_args = {}
        self.default_schema_file = Path(__file__).parent.joinpath(
            "data/CfnLintCli/config/schema.json"
        )
        with self.default_schema_file.open(encoding="utf-8") as f:
            self.default_schema = json.load(f)
        self.schema = self.default_schema if not schema else schema

        if config_file:
            self._custom_config_file = config_file
        else:
            LOGGER.debug("Looking for CFLINTRC before attempting to load")
            self._user_config_file, self._project_config_file = self._find_config()

        self.load()

    def _find_config(self):
        """Looks up for user and project level config
        Returns
        -------
        Tuple
            (Path, Path)
            Tuple with both configs and whether they were found
        Example
        -------
            > user_config, project_config = self._find_config()
        """
        config_file_name = ".cfnlintrc"

        user_config_path = ""
        home_path = Path.home()
        for path in [
            home_path.joinpath(config_file_name),
            home_path.joinpath(f"{config_file_name}.yaml"),
            home_path.joinpath(f"{config_file_name}.yml"),
        ]:
            if self._has_file(path):
                user_config_path = path
                break

        project_config_path = ""
        cwd_path = Path.cwd()
        for path in [
            cwd_path.joinpath(config_file_name),
            cwd_path.joinpath(f"{config_file_name}.yaml"),
            cwd_path.joinpath(f"{config_file_name}.yml"),
        ]:
            if self._has_file(path):
                project_config_path = path
                break

        return user_config_path, project_config_path

    def _has_file(self, filename):
        """Confirm whether file exists
        Parameters
        ----------
        filename : str
            Path to a file
        Returns
        -------
        Boolean
        """

        return Path(filename).is_file()

    def load(self):
        """Load configuration file and expose as a dictionary
        Returns
        -------
        Dict
            CFLINTRC configuration
        """

        if self._custom_config_file:
            custom_config = self._read_config(self._custom_config_file)
            LOGGER.debug("Validating Custom CFNLINTRC")
            self.validate_config(custom_config, self.schema)
            LOGGER.debug("Custom configuration loaded as")
            LOGGER.debug("%s", custom_config)

            self.file_args = custom_config
        else:
            user_config = self._read_config(self._user_config_file)
            LOGGER.debug("Validating User CFNLINTRC")
            self.validate_config(user_config, self.schema)

            project_config = self._read_config(self._project_config_file)
            LOGGER.debug("Validating Project CFNLINTRC")
            self.validate_config(project_config, self.schema)

            LOGGER.debug("User configuration loaded as")
            LOGGER.debug("%s", user_config)
            LOGGER.debug("Project configuration loaded as")
            LOGGER.debug("%s", project_config)

            LOGGER.debug("Merging configurations...")
            self.file_args = self.merge_config(user_config, project_config)

    def validate_config(self, config, schema):
        """Validate configuration against schema
        Parameters
        ----------
        config : dict
            CFNLINTRC configuration
        schema : dict
            JSONSchema to validate against
        Raises
        -------
        jsonschema.exceptions.ValidationError
            Returned when cfnlintrc doesn't match schema provided
        """
        LOGGER.debug("Validating CFNLINTRC config with given JSONSchema")
        LOGGER.debug("Schema used: %s", schema)
        LOGGER.debug("Config used: %s", config)

        validator = StandardValidator(schema=schema)
        validator.validate(config)
        LOGGER.debug("CFNLINTRC looks valid!")

    def merge_config(self, user_config, project_config):
        """Merge project and user configuration into a single dictionary
        Creates a new configuration with both configuration merged
        it favours project level over user configuration if keys are duplicated
        NOTE
        ----
            It takes any number of nested dicts
            It overrides lists found in user_config with project_config
        Parameters
        ----------
        user_config : Dict
            User configuration (~/.cfnlintrc) found at user's home directory
        project_config : Dict
            Project configuration (.cfnlintrc) found at current directory
        Returns
        -------
        Dict
            Merged configuration
        """
        # Recursively override User config with Project config
        for key in user_config:
            if key in project_config:
                # If both keys are the same, let's check whether they have nested keys
                if isinstance(user_config[key], dict) and isinstance(
                    project_config[key], dict
                ):
                    self.merge_config(user_config[key], project_config[key])
                else:
                    user_config[key] = project_config[key]
                    LOGGER.debug(
                        "Overriding User's key %s with Project's specific value %s.",
                        key,
                        project_config[key],
                    )

        # Project may have unique config we need to copy over too
        # so that we can have user+project config available as one
        for key in project_config:
            if key not in user_config:
                user_config[key] = project_config[key]

        return user_config

    def _read_config(self, config):
        """Parse given YAML configuration
        Returns
        -------
        Dict
            Parsed YAML configuration as dictionary
        """
        config = Path(config)
        config_template = None

        if self._has_file(config):
            LOGGER.debug("Parsing CFNLINTRC")
            config_template = cfnlint.decode.cfn_yaml.load(str(config))

        if not config_template:
            config_template = {}

        return config_template


def comma_separated_arg(string):
    """Split a comma separated string"""
    return string.split(",")


def _ensure_value(namespace, name, value):
    if getattr(namespace, name, None) is None:
        setattr(namespace, name, value)
    return getattr(namespace, name)


class RuleConfigurationAction(argparse.Action):
    """Override the default Action"""

    def __init__(
        self,
        option_strings,
        dest,
        nargs=None,
        const=None,
        default=None,
        type=None,
        choices=None,
        required=False,
        help=None,
        metavar=None,
    ):  # pylint: disable=W0622
        super().__init__(
            option_strings=option_strings,
            dest=dest,
            nargs=nargs,
            const=const,
            default=default,
            type=type,
            choices=choices,
            required=required,
            help=help,
            metavar=metavar,
        )

    def _parse_rule_configuration(self, string):
        """Parse the config rule structure"""
        configs = comma_separated_arg(string)
        results = {}
        for config in configs:
            rule_id = config.split(":")[0]
            config_name = config.split(":")[1].split("=")[0]
            config_value = config.split(":")[1].split("=")[1]
            if rule_id not in results:
                results[rule_id] = {}
            results[rule_id][config_name] = config_value

        return results

    def __call__(self, parser, namespace, values, option_string=None):
        items = copy.copy(_ensure_value(namespace, self.dest, {}))
        try:
            for value in values:
                new_value = self._parse_rule_configuration(value)
                for v_k, v_vs in new_value.items():
                    if v_k in items:
                        for s_k, s_v in v_vs.items():
                            items[v_k][s_k] = s_v
                    else:
                        items[v_k] = v_vs
            setattr(namespace, self.dest, items)
        except Exception:  # pylint: disable=W0703
            parser.print_help()
            parser.exit(1)


class CliArgs:
    """Base Args class"""

    def __init__(self, cli_args: Sequence[str] | None):
        self.parser = self.create_parser()
        self.cli_args = self.parser.parse_args(cli_args or [])

    def create_parser(self):
        """Do first round of parsing parameters to set options"""

        class ArgumentParser(argparse.ArgumentParser):
            """Override Argument Parser so we can control the exit code"""

            def error(self, message):
                self.print_help(sys.stderr)
                self.exit(1, f"{self.prog}: error: {message}\n")

        class ExtendAction(argparse.Action):
            """Support argument types that are lists and can
            be specified multiple times.
            """

            def __call__(self, parser, namespace, values, option_string=None):
                items = getattr(namespace, self.dest)
                items = [] if items is None else items
                for value in values:
                    if isinstance(value, list):
                        items.extend(value)
                    else:
                        items.append(value)
                setattr(namespace, self.dest, items)

        usage = (
            "\nBasic: cfn-lint test.yaml\n"
            "Ignore a rule: cfn-lint -i E3012 -- test.yaml\n"
            "Configure a rule: cfn-lint -x E3012:strict=true -t test.yaml\n"
            "Lint all yaml files in a folder: cfn-lint dir/**/*.yaml"
        )

        parser = ArgumentParser(description="CloudFormation Linter", usage=usage)
        parser.register("action", "extend", ExtendAction)

        standard = parser.add_argument_group("Standard")
        advanced = parser.add_argument_group("Advanced / Debugging")

        # Allow the template to be passes as an optional or a positional argument
        standard.add_argument(
            "templates",
            metavar="TEMPLATE",
            nargs="*",
            help="The CloudFormation template to be linted",
        )
        standard.add_argument(
            "-t",
            "--template",
            metavar="TEMPLATE",
            dest="template_alt",
            help="The CloudFormation template to be linted",
            nargs="+",
            default=[],
            action="extend",
        )
        standard.add_argument(
            "-b",
            "--ignore-bad-template",
            help="Ignore failures with Bad template",
            action="store_true",
        )
        standard.add_argument(
            "--ignore-templates",
            dest="ignore_templates",
            help="Ignore templates",
            nargs="+",
            default=[],
            action="extend",
        )
        advanced.add_argument(
            "-D", "--debug", help="Enable debug logging", action="store_true"
        )
        advanced.add_argument(
            "-I", "--info", help="Enable information logging", action="store_true"
        )
        standard.add_argument(
            "-f",
            "--format",
            help="Output Format",
            choices=["quiet", "parseable", "json", "junit", "pretty", "sarif"],
        )
        standard.add_argument(
            "-l",
            "--list-rules",
            dest="listrules",
            default=False,
            action="store_true",
            help="list all the rules",
        )
        standard.add_argument(
            "-r",
            "--regions",
            dest="regions",
            nargs="+",
            default=[],
            type=comma_separated_arg,
            action="extend",
            help="list the regions to validate against.",
        )
        advanced.add_argument(
            "-a",
            "--append-rules",
            dest="append_rules",
            nargs="+",
            default=[],
            type=comma_separated_arg,
            action="extend",
            help=(
                "specify one or more rules directories using "
                "one or more --append-rules arguments. "
            ),
        )
        standard.add_argument(
            "-i",
            "--ignore-checks",
            dest="ignore_checks",
            nargs="+",
            default=[],
            type=comma_separated_arg,
            action="extend",
            help="only check rules whose id do not match these values",
        )
        standard.add_argument(
            "-c",
            "--include-checks",
            dest="include_checks",
            nargs="+",
            default=[],
            type=comma_separated_arg,
            action="extend",
            help="include rules whose id match these values",
        )
        standard.add_argument(
            "-m",
            "--mandatory-checks",
            dest="mandatory_checks",
            nargs="+",
            default=[],
            type=comma_separated_arg,
            action="extend",
            help=(
                "always check rules whose id match these values, regardless of template"
                " exclusions"
            ),
        )
        standard.add_argument(
            "-e",
            "--include-experimental",
            help="Include experimental rules",
            action="store_true",
        )
        standard.add_argument(
            "-x",
            "--configure-rule",
            dest="configure_rules",
            nargs="+",
            default={},
            action=RuleConfigurationAction,
            help=(
                "Provide configuration for a rule. Format RuleId:key=value. Example:"
                " E3012:strict=true"
            ),
        )
        standard.add_argument(
            "--config-file",
            dest="config_file",
            help="Specify the cfnlintrc file to use",
        )
        standard.add_argument(
            "-z",
            "--custom-rules",
            dest="custom_rules",
            help="Allows specification of a custom rule file.",
        )
        advanced.add_argument(
            "-o",
            "--override-spec",
            dest="override_spec",
            help="A CloudFormation Spec override file that allows customization",
        )
        advanced.add_argument(
            "-g",
            "--build-graph",
            help=(
                "Creates a file in the same directory as the template that models the"
                " template's resources in DOT format"
            ),
            action="store_true",
        )
        advanced.add_argument(
            "-s",
            "--registry-schemas",
            default=[],
            help="one or more directories of CloudFormation Registry Schemas",
            action="extend",
            type=comma_separated_arg,
            nargs="+",
        )
        standard.add_argument(
            "-v",
            "--version",
            help="Version of cfn-lint",
            action="version",
            version=f"%(prog)s {__version__}",
        )
        advanced.add_argument(
            "-u",
            "--update-specs",
            help="Update the CloudFormation Specs",
            action="store_true",
        )
        advanced.add_argument(
            "-p",
            "--patch-specs",
            help="Patch the CloudFormation Specs in place",
            action="store_true",
        )
        advanced.add_argument(
            "--update-documentation", help=argparse.SUPPRESS, action="store_true"
        )
        advanced.add_argument(
            "--update-iam-policies", help=argparse.SUPPRESS, action="store_true"
        )
        standard.add_argument(
            "--output-file",
            type=str,
            default=None,
            help="Writes the output to the specified file, ideal for producing reports",
        )
        standard.add_argument(
            "--merge-configs",
            default=False,
            action="store_true",
            help="Merges lists between configuration layers",
        )
        standard.add_argument(
            "--non-zero-exit-code",
            type=str,
            default="informational",
            choices=["informational", "warning", "error", "none"],
            help="Exit code will be non zero from the specified rule class and higher",
        )
        advanced.add_argument("--force", help=argparse.SUPPRESS, action="store_true")

        return parser


class TemplateArgs:
    """Per Template Args"""

    def __init__(self, template_args):
        self.set_template_args(template_args)

    def get_template_args(self):
        return self._template_args

    def set_template_args(self, template):
        defaults = {}
        if isinstance(template, dict):
            configs = template.get("Metadata", {}).get("cfn-lint", {}).get("config", {})

            if isinstance(configs, dict):
                for key, value in {
                    "ignore_checks": (list),
                    "regions": (list),
                    "append_rules": (list),
                    "override_spec": (str),
                    "custom_rules": (str),
                    "ignore_bad_template": (bool),
                    "include_checks": (list),
                    "configure_rules": (dict),
                    "include_experimental": (bool),
                }.items():
                    if key in configs:
                        if isinstance(configs[key], value):
                            defaults[key] = configs[key]

        self._template_args = defaults

    template_args = property(get_template_args, set_template_args)


class ManualArgs(TypedDict, total=False):
    configure_rules: dict[str, dict[str, Any]]
    include_checks: list[str]
    ignore_checks: list[str]
    mandatory_checks: list[str]
    include_experimental: bool
    ignore_bad_template: bool
    ignore_templates: list
    merge_configs: bool
    non_zero_exit_code: str
    output_file: str
    regions: list
    registry_schemas: list[str]


def _merge_configs(
    cli_value: Any, template_value: Any, file_value: Any, manual_value: Any
) -> Any:
    # the CLI will always have an empty list when the item is a list
    # we will use that to evaluate if we need to merge the lists
    if isinstance(cli_value, list):
        merged_list = cli_value.copy()
        if isinstance(template_value, list):
            merged_list.extend(template_value)
        if isinstance(file_value, list):
            merged_list.extend(file_value)
        if isinstance(manual_value, list):
            merged_list.extend(manual_value)
        return merged_list

    elif isinstance(cli_value, dict):
        merged_dict = cli_value.copy()
        if isinstance(template_value, dict):
            merged_dict.update(template_value)
        if isinstance(file_value, dict):
            merged_dict.update(file_value)
        if isinstance(manual_value, dict):
            merged_dict.update(manual_value)
        return merged_dict

    return None


# pylint: disable=too-many-public-methods
class ConfigMixIn(TemplateArgs, CliArgs, ConfigFileArgs):

    def __init__(self, cli_args: list[str] | None = None, **kwargs: Unpack[ManualArgs]):
        self._manual_args = kwargs or ManualArgs()
        self._templates_to_process = False
        CliArgs.__init__(self, cli_args)
        # configure debug as soon as we can
        TemplateArgs.__init__(self, {})
        ConfigFileArgs.__init__(
            self, config_file=self._get_argument_value("config_file", False, False)
        )

    def __repr__(self):
        return format_json_string(
            {
                "ignore_checks": self.ignore_checks,
                "include_checks": self.include_checks,
                "mandatory_checks": self.mandatory_checks,
                "include_experimental": self.include_experimental,
                "configure_rules": self.configure_rules,
                "registry_schemas": self.registry_schemas,
                "regions": self.regions,
                "ignore_bad_template": self.ignore_bad_template,
                "debug": self.debug,
                "info": self.info,
                "format": self.format,
                "templates": self.templates,
                "append_rules": self.append_rules,
                "override_spec": self.override_spec,
                "custom_rules": self.custom_rules,
                "config_file": self.config_file,
                "merge_configs": self.merge_configs,
                "non_zero_exit_code": self.non_zero_exit_code,
                "patch_specs": self.patch_specs,
            }
        )

    def _get_argument_value(self, arg_name, is_template, is_config_file):
        cli_value = getattr(self.cli_args, arg_name)
        template_value = self.template_args.get(arg_name)
        file_value = self.file_args.get(arg_name)
        manual_value = self._manual_args.get(arg_name)

        # merge list configurations
        # make sure we don't do an infinite loop so skip this check for merge_configs
        if arg_name != "merge_configs":
            if self.merge_configs:
                if isinstance(cli_value, (list, dict)):
                    return _merge_configs(
                        cli_value, template_value, file_value, manual_value
                    )

        # return individual items
        if manual_value:
            return manual_value
        if cli_value:
            return cli_value
        if template_value and is_template:
            return template_value
        if file_value and is_config_file:
            return file_value
        return cli_value

    @property
    def ignore_checks(self):
        return self._get_argument_value("ignore_checks", True, True)

    @property
    def include_checks(self):
        results = self._get_argument_value("include_checks", True, True)
        return ["W", "E"] + results

    @property
    def mandatory_checks(self):
        return self._get_argument_value("mandatory_checks", False, True)

    @property
    def include_experimental(self):
        return self._get_argument_value("include_experimental", True, True)

    @property
    def regions(self):
        results = self._get_argument_value("regions", True, True)
        if not results:
            default_region_env = os.environ.get("AWS_DEFAULT_REGION", "us-east-1")
            return [os.environ.get("AWS_REGION", default_region_env)]
        if "ALL_REGIONS" in results:
            return REGIONS
        return results

    @property
    def ignore_bad_template(self):
        return self._get_argument_value("ignore_bad_template", True, True)

    @property
    def debug(self):
        return self._get_argument_value("debug", False, True)

    @property
    def info(self):
        return self._get_argument_value("info", False, False)

    @property
    def format(self):
        return self._get_argument_value("format", False, True)

    @property
    def templates(self):
        """

        Returns a list of Cloudformation templates to lint.

        Order of precedence:
        - Filenames provided via `-t` CLI
        - Filenames specified in the config file.
        - Arguments provided via `cfn-lint` CLI.
        """

        all_filenames = []

        cli_alt_args = self._get_argument_value("template_alt", False, False)
        file_args = self._get_argument_value("templates", False, True)
        cli_args = self._get_argument_value("templates", False, False)

        if cli_alt_args:
            filenames = cli_alt_args
        elif cli_args:
            filenames = cli_args
        # elif not sys.stdin.isatty():
        #    if (select.select([sys.stdin], [], [], 0.0)[0]):
        #        return []
        elif file_args:
            filenames = file_args
        else:
            # No filenames found, could be piped in or be using the api.
            return None

        # If we're still haven't returned, we've got templates to lint.
        # Build up list of templates to lint.
        self.templates_to_process = True

        if isinstance(filenames, str):
            filenames = [filenames]

        ignore_templates = self._ignore_templates()
        all_filenames.extend(self._glob_filenames(filenames))

        found_files = [i for i in all_filenames if i not in ignore_templates]
        LOGGER.debug(
            f"List of Cloudformation Templates to lint: {found_files} from {filenames}"
        )
        return found_files

    def _ignore_templates(self):
        ignore_template_args = self._get_argument_value("ignore_templates", False, True)
        if ignore_template_args:
            filenames = ignore_template_args
        else:
            return []

        # if only one is specified convert it to array
        if isinstance(filenames, str):
            filenames = [filenames]

        return self._glob_filenames(filenames, False)

    def _glob_filenames(
        self, filenames: Sequence[str], raise_exception: bool = True
    ) -> list[str]:
        # handle different shells and Config files
        # some shells don't expand * and configparser won't expand wildcards
        all_filenames = []

        for filename in filenames:
            add_filenames = glob.glob(filename, recursive=True)

            if not add_filenames and not self.ignore_bad_template:
                if raise_exception:
                    raise ValueError(f"{filename} could not be processed by glob.glob")

            all_filenames.extend(add_filenames)

        return sorted(list(map(str, map(Path, all_filenames))))

    @property
    def append_rules(self):
        return [_DEFAULT_RULESDIR] + self._get_argument_value(
            "append_rules", False, True
        )

    @property
    def override_spec(self):
        return self._get_argument_value("override_spec", False, True)

    @property
    def custom_rules(self):
        """custom_rules_spec"""
        return self._get_argument_value("custom_rules", False, True)

    @property
    def update_specs(self):
        return self._get_argument_value("update_specs", False, False)

    @property
    def patch_specs(self):
        return self._get_argument_value("patch_specs", False, False)

    @property
    def update_documentation(self):
        return self._get_argument_value("update_documentation", False, False)

    @property
    def update_iam_policies(self):
        return self._get_argument_value("update_iam_policies", False, False)

    @property
    def listrules(self):
        return self._get_argument_value("listrules", False, False)

    @property
    def configure_rules(self):
        return self._get_argument_value("configure_rules", True, True)

    @property
    def config_file(self):
        return self._get_argument_value("config_file", False, False)

    @property
    def build_graph(self):
        return self._get_argument_value("build_graph", False, False)

    @property
    def output_file(self):
        return self._get_argument_value("output_file", False, True)

    @property
    def registry_schemas(self):
        return self._get_argument_value("registry_schemas", False, True)

    @property
    def merge_configs(self):
        return self._get_argument_value("merge_configs", True, True)

    @property
    def non_zero_exit_code(self):
        return self._get_argument_value("non_zero_exit_code", False, False)

    @property
    def force(self):
        return self._get_argument_value("force", False, False)

    @property
    def templates_to_process(self):
        return self._templates_to_process

    @templates_to_process.setter
    def templates_to_process(self, value: bool):
        self._templates_to_process = value



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import json
import sys
from pathlib import Path
from test.testlib.testcase import BaseTestCase
from unittest.mock import patch

import defusedxml.ElementTree as ET

import cfnlint.config
import cfnlint.formatters
import cfnlint.helpers
from cfnlint import ConfigMixIn
from cfnlint.formatters import (
    Formatter,
    JsonFormatter,
    JUnitFormatter,
    ParseableFormatter,
    PrettyFormatter,
    QuietFormatter,
    SARIFFormatter,
)
from cfnlint.jsonschema import StandardValidator
from cfnlint.rules import Match, Rules
from cfnlint.rules.functions.SubUnneeded import SubUnneeded
from cfnlint.rules.resources.properties.Type import Type
from cfnlint.rules.resources.UpdateReplacePolicyDeletionPolicyOnStatefulResourceTypes import (
    UpdateReplacePolicyDeletionPolicyOnStatefulResourceTypes,
)


class TestFormatters(BaseTestCase):
    """Test Formatters"""

    def setUp(self) -> None:
        super().setUp()
        self.rules = Rules.create_from_directory(cfnlint.config._DEFAULT_RULESDIR)
        self.filename = str(Path("test/fixtures/templates/bad/formatters.yaml"))
        self.sarif_schema = str(Path("test/fixtures/schemas/sarif/schema-2.1.0.json"))
        self.config = ConfigMixIn(
            cli_args=[
                "--include-checks",
                "I",
                "--ignore-checks",
                "E1029",
                "--",
                self.filename,
            ]
        )

        self.results = [
            Match(
                linenumber=6,
                columnnumber=3,
                linenumberend=6,
                columnnumberend=10,
                filename=self.filename,
                rule=UpdateReplacePolicyDeletionPolicyOnStatefulResourceTypes(),
                message=(
                    "The default action when replacing/removing a resource is to delete"
                    " it. Set explicit values for UpdateReplacePolicy / DeletionPolicy"
                    " on potentially stateful resource: Resources/myTable"
                ),
            ),
            Match(
                linenumber=9,
                columnnumber=7,
                linenumberend=9,
                columnnumberend=16,
                filename=self.filename,
                rule=SubUnneeded(),
                message=(
                    "Fn::Sub isn't needed because there are no variables at"
                    " Resources/myTable/Properties/TableName/Fn::Sub"
                ),
            ),
            Match(
                linenumber=18,
                columnnumber=9,
                linenumberend=18,
                columnnumberend=27,
                filename=self.filename,
                rule=Type(),
                message="'5' is not of type 'integer'",
            ),
        ]

    def test_base_formatter(self):
        """Test base formatter"""
        formatter = Formatter()
        a_results = formatter.print_matches(
            self.results, self.rules, self.config
        ).splitlines()

        # Check the errors
        self.assertEqual(
            a_results[0], f"{self.results[0].rule.id} {self.results[0].message}"
        )
        self.assertEqual(
            a_results[1],
            f"{self.results[0].filename}:{self.results[0].linenumber}:{self.results[0].columnnumber}",
        )
        self.assertEqual(a_results[2], "")
        self.assertEqual(
            a_results[3], f"{self.results[1].rule.id} {self.results[1].message}"
        )
        self.assertEqual(
            a_results[4],
            f"{self.results[1].filename}:{self.results[1].linenumber}:{self.results[1].columnnumber}",
        )
        self.assertEqual(a_results[5], "")
        self.assertEqual(
            a_results[6], f"{self.results[2].rule.id} {self.results[2].message}"
        )
        self.assertEqual(
            a_results[7],
            f"{self.results[2].filename}:{self.results[2].linenumber}:{self.results[2].columnnumber}",
        )

    def test_quiet_formatter(self):
        """Test quiet formatter"""

        formatter = QuietFormatter()
        a_results = formatter.print_matches(
            self.results, self.rules, self.config
        ).splitlines()

        for i in range(3):
            # Check the errors
            # ruff: noqa: E501
            self.assertEqual(
                a_results[i],
                (
                    f"{self.results[i].rule.id}:"
                    f" {self.results[i].rule.shortdesc} {self.results[i].filename}:{self.results[i].linenumber}"
                ),
            )

    def test_parseable_formatter(self):
        """Test Parseable formatter"""
        formatter = ParseableFormatter()
        results = formatter.print_matches(
            self.results, self.rules, self.config
        ).splitlines()

        for i in range(3):
            # Check the errors
            self.assertEqual(
                results[i],
                f"{self.results[i].filename}:{self.results[i].linenumber}:{self.results[i].columnnumber}:{self.results[i].linenumberend}:{self.results[i].columnnumberend}:{self.results[i].rule.id}:{self.results[i].message}",
            )

    def test_pretty_formatter(self):
        """Test pretty formatter"""
        formatter = PrettyFormatter()
        results = formatter.print_matches(
            self.results, rules=self.rules, config=self.config
        ).splitlines()

        if sys.stdout.isatty():
            # Check the errors
            self.assertEqual(results[0], f"\x1b[4m{self.filename}\x1b[0m")
            self.assertEqual(
                results[1],
                (
                    f"\x1b[0m{self.results[2].linenumber}:{self.results[2].columnnumber}:"
                    f"               \x1b[0m\x1b[31m{self.results[2].rule.id}    "
                    f" \x1b[0m{self.results[2].message}"
                ),
            )
            self.assertEqual(
                results[2],
                (
                    f"\x1b[0m{self.results[1].linenumber}:{self.results[1].columnnumber}:"
                    f"                \x1b[0m\x1b[33m{self.results[1].rule.id}    "
                    f" \x1b[0m{self.results[1].message}"
                ),
            )
            self.assertEqual(
                results[3],
                (
                    f"\x1b[0m{self.results[0].linenumber}:{self.results[0].columnnumber}:"
                    f"                \x1b[0m\x1b[34m{self.results[0].rule.id}    "
                    f" \x1b[0m{self.results[0].message}"
                ),
            )
        else:
            # Check the errors
            self.assertEqual(results[0], self.filename)
            self.assertEqual(
                results[1],
                (
                    f"{self.results[2].linenumber}:{self.results[2].columnnumber}:     "
                    f"          {self.results[2].rule.id}     {self.results[2].message}"
                ),
            )
            self.assertEqual(
                results[2],
                (
                    f"{self.results[1].linenumber}:{self.results[1].columnnumber}:     "
                    f"           {self.results[1].rule.id}    "
                    f" {self.results[1].message}"
                ),
            )
            self.assertEqual(
                results[3],
                (
                    f"{self.results[0].linenumber}:{self.results[0].columnnumber}:     "
                    f"           {self.results[0].rule.id}    "
                    f" {self.results[0].message}"
                ),
            )

    @patch("sys.stdin.isatty", return_va=True)
    def test_pretty_formatter_pipe(self, isatty_mock):
        """Test pretty formatter"""
        isatty_mock.return_value = True
        formatter = PrettyFormatter()
        self.config.cli_args.templates = None
        results = formatter.print_matches(
            self.results, rules=self.rules, config=self.config
        ).splitlines()

        if sys.stdout.isatty():
            self.assertIn("Cfn-lint scanned 1 templates", results[5])
        else:
            # Check the errors
            self.assertIn("Cfn-lint scanned 1 templates", results[5])

    def test_json_formatter(self):
        """Test JSON formatter"""
        formatter = JsonFormatter()

        # Get the JSON output
        json_results = json.loads(
            formatter.print_matches(self.results, self.rules, self.config)
        )

        # Check the 3 errors again
        self.assertEqual(len(json_results), 3)

        # Check the errors level description
        for i in range(3):
            self.assertEqual(json_results[i]["Rule"]["Id"], self.results[i].rule.id)
            self.assertEqual(json_results[i]["Message"], self.results[i].message)

    def test_junit_returns_none(self):
        """Test JUnut Formatter returns None if no rules are passed in"""
        formatter = JUnitFormatter()

        # The actual test
        self.assertIsNone(formatter.print_matches([], [], self.config))

    def test_junit_formatter(self):
        """Test JUnit Formatter"""
        formatter = JUnitFormatter()

        self.rules._used_rules = {
            "I3011": self.rules["I3011"],
            "E3012": self.rules["E3012"],
            "W1020": self.rules["W1020"],
        }
        s = formatter.print_matches(self.results, self.rules, self.config)
        root = ET.fromstring(s)

        self.assertEqual(root.tag, "testsuites")
        self.assertEqual(root[0].tag, "testsuite")

        found_i3011 = False
        found_w1020 = False
        found_e3012 = False
        name_i3011 = "{0} {1}".format(
            self.results[0].rule.id, self.results[0].rule.shortdesc
        )
        name_w1020 = "{0} {1}".format(
            self.results[1].rule.id, self.results[1].rule.shortdesc
        )
        name_e3012 = "{0} {1}".format(
            self.results[2].rule.id, self.results[2].rule.shortdesc
        )
        name_e1029 = "E1029 Sub is required if a variable is used in a string"
        for child in root[0]:
            self.assertEqual(child.tag, "testcase")

            if child.attrib["name"] in [name_i3011, name_w1020, name_e3012]:
                self.assertEqual(child[0].tag, "failure")

                if child.attrib["name"] == name_i3011:
                    found_i3011 = True
                if child.attrib["name"] == name_w1020:
                    found_w1020 = True
                if child.attrib["name"] == name_e3012:
                    found_e3012 = True

            if child.attrib["name"] == name_e1029:
                self.assertEqual(child[0].tag, "skipped")
                self.assertEqual(child[0].attrib["type"], "skipped")

        self.assertTrue(found_i3011)
        self.assertTrue(found_w1020)
        self.assertTrue(found_e3012)

    def test_sarif_formatter(self):
        """Test the SARIF formatter"""
        formatter = SARIFFormatter()

        # Get the SARIF JSON output
        sarif = json.loads(
            formatter.print_matches(self.results, self.rules, self.config)
        )

        with open(self.sarif_schema, encoding="utf-8") as f:
            schema = json.load(f)
        # Fetch the SARIF schema
        # schema = json.loads(cfnlint.helpers.get_url_content(sarif["$schema"], False))
        validator = StandardValidator(schema=schema)
        validator.validate(sarif)

        sarif_results = sarif["runs"][0]["results"]
        # Check the 3 errors again
        self.assertEqual(len(sarif_results), 3)

        # Sanity check the errors
        self.assertEqual(sarif_results[0]["level"], "note")
        self.assertEqual(sarif_results[0]["ruleId"], "I3011")
        # IMPORTANT: 'warning' is the default level in SARIF (when kind is
        # absent) and is stripped by serialization
        self.assertNotIn("level", sarif_results[1].keys())
        self.assertEqual(sarif_results[1]["ruleId"], "W1020")
        self.assertEqual(sarif_results[2]["level"], "error")
        self.assertEqual(sarif_results[2]["ruleId"], "E3012")



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.testlib.testcase import BaseTestCase

from cfnlint.context import Resource
from cfnlint.template.getatts import GetAtts


class TestGetAtts(BaseTestCase):
    """Test GetAtts Class in template"""

    def test_getatt(self):
        getatts = GetAtts(["us-east-1", "us-west-2"])
        getatts.add(
            "Module", Resource({"Type": "Organization::Resource::Type::MODULE"})
        )
        results = getatts.match("us-east-1", "ModuleResource.Module")
        self.assertEqual(results, "/properties/CfnLintAllTypes")

    def test_many_modules(self):
        getatts = GetAtts(["us-east-1", "us-west-2"])
        getatts.add(
            "Module", Resource({"Type": "Organization::Resource::Type::MODULE"})
        )
        getatts.add(
            "Module1", Resource({"Type": "Organization::Resource::Type::MODULE"})
        )
        results = getatts.match("us-east-1", "ModuleResource.Module")
        self.assertEqual(results, "/properties/CfnLintAllTypes")

        results = getatts.match("us-east-1", "Module1Resource.Module")
        self.assertEqual(results, "/properties/CfnLintAllTypes")

    def test_getatt_resource_and_modules(self):
        getatts = GetAtts(["us-east-1", "us-west-2"])
        getatts.add(
            "Resource", Resource({"Type": "Organization::Resource::Type::MODULE"})
        )
        getatts.add("Resource1", Resource({"Type": "Organization::Resource::Type"}))
        results = getatts.match("us-east-1", "ResourceOne.Module")
        self.assertEqual(results, "/properties/CfnLintAllTypes")

        with self.assertRaises(ValueError):
            getatts.match("us-east-1", "NoMatch.Attribute")

    def test_getatt_type_errors(self):
        getatts = GetAtts(["us-east-1"])

        with self.assertRaises(TypeError):
            getatts.match("us-east-1", {})

        with self.assertRaises(TypeError):
            getatts.match("us-east-1", ["Foo", "Bar", "FooBar"])

    def test_getatt_resource_with_list(self):
        getatts = GetAtts(["us-east-1"])
        getatts.add("Resource", Resource({"Type": "AWS::NetworkFirewall::Firewall"}))
        results = getatts.match("us-east-1", "Resource.EndpointIds")
        self.assertEqual(results, "/properties/EndpointIds")
        self.assertDictEqual(
            getatts.json_schema("us-east-1"),
            {
                "oneOf": [
                    {
                        "type": "array",
                        "items": [
                            {"type": "string", "enum": ["Resource"]},
                            {"type": ["string", "object"]},
                        ],
                        "allOf": [
                            {
                                "if": {
                                    "items": [
                                        {"type": "string", "const": "Resource"},
                                        {"type": ["string", "object"]},
                                    ]
                                },
                                "then": {
                                    "if": {
                                        "items": [
                                            {"type": "string", "const": "Resource"},
                                            {"type": "string"},
                                        ]
                                    },
                                    "then": {
                                        "items": [
                                            {"type": "string", "const": "Resource"},
                                            {
                                                "type": "string",
                                                "enum": [
                                                    "FirewallArn",
                                                    "FirewallId",
                                                    "EndpointIds",
                                                ],
                                            },
                                        ]
                                    },
                                    "else": {
                                        "items": [
                                            {"type": "string", "const": "Resource"},
                                            {
                                                "type": "object",
                                                "properties": {
                                                    "Ref": {"type": "string"}
                                                },
                                                "required": ["Ref"],
                                                "additionalProperties": False,
                                            },
                                        ]
                                    },
                                },
                                "else": {},
                            }
                        ],
                    },
                    {
                        "type": "string",
                        "enum": [
                            "Resource.FirewallArn",
                            "Resource.FirewallId",
                            "Resource.EndpointIds",
                        ],
                    },
                ]
            },
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import unicode_literals

from cfnlint._typing import RuleMatches
from cfnlint.rules import CloudFormationLintRule, RuleMatch
from cfnlint.template import Template


class Used(CloudFormationLintRule):
    """Check if Parameters are used"""

    id = "W2001"
    shortdesc = "Check if Parameters are Used"
    description = "Making sure the parameters defined are used"
    source_url = "https://github.com/aws-cloudformation/cfn-lint"
    tags = ["parameters"]

    def match(self, cfn: Template) -> RuleMatches:
        matches: RuleMatches = []
        if cfn.transform_pre["Transform"]:
            return matches

        le_refs = None
        if cfn.has_language_extensions_transform():
            le_refs = cfn.search_deep_keys("Ref")

        reftrees = cfn.transform_pre.get("Ref", [])
        subtrees = cfn.transform_pre.get("Fn::Sub", [])
        refs = []
        for reftree in reftrees:
            refs.append(reftree[-1])
        if le_refs:
            for le_ref in le_refs:
                if le_ref[-1] not in refs:
                    refs.append(le_ref[-1])

        subs = []
        for subtree in subtrees:
            if isinstance(subtree[-1], list):
                subs.extend(cfn.get_sub_parameters(subtree[-1][0]))
            elif isinstance(subtree[-1], str):
                subs.extend(cfn.get_sub_parameters(subtree[-1]))

        for paramname in cfn.template.get("Parameters", {}).keys():
            if paramname not in refs:
                if paramname not in subs:
                    message = "Parameter {0} not used."
                    matches.append(
                        RuleMatch(["Parameters", paramname], message.format(paramname))
                    )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import CfnTemplateValidator, ValidationError
from cfnlint.rules.resources.codepipeline.PipelineActionConfiguration import (
    PipelineActionConfiguration,
)


@pytest.fixture
def rule():
    rule = PipelineActionConfiguration()
    yield rule


def format(validator, keywords, instance, schema):
    if instance:
        return

    yield ValidationError("bad")


@pytest.fixture
def validator(cfn, context):
    validator = CfnTemplateValidator({}).extend(
        validators={
            "format": format,
        }
    )

    return validator(
        context=context,
        cfn=cfn,
        schema={},
    )


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            [],
            [],
        ),
        (
            {
                "Configuration": {
                    "TemplatePath": "Source::template.yaml",
                },
                "InputArtifacts": [{"Name": "Source"}],
            },
            [],
        ),
        (
            {
                "Configuration": {
                    "TemplatePath": "Foo::template.yaml",
                },
                "InputArtifacts": [{"Name": "Bar"}],
            },
            [
                ValidationError(
                    "'Foo' is not one of ['Bar']",
                    rule=PipelineActionConfiguration(),
                    validator="enum",
                    path_override=deque(["Configuration", "TemplatePath"]),
                    schema_path=deque([]),
                )
            ],
        ),
        (
            {
                "Configuration": {
                    "TemplatePath": "Foo::template.yaml",
                },
            },
            [
                ValidationError(
                    "'Foo' is not one of []",
                    rule=PipelineActionConfiguration(),
                    validator="enum",
                    path_override=deque(["Configuration", "TemplatePath"]),
                    schema_path=deque([]),
                )
            ],
        ),
        (
            {
                "Configuration": {
                    "RoleArn": True,
                },
            },
            [],
        ),
        (
            {
                "Configuration": {
                    "RoleArn": False,
                },
            },
            [
                ValidationError(
                    "bad",
                    rule=PipelineActionConfiguration(),
                    validator="format",
                    path_override=deque(["Configuration", "RoleArn"]),
                    schema_path=deque(["format"]),
                )
            ],
        ),
    ],
)
def test_validate(instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))
    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from typing import Any

from cfnlint.jsonschema.protocols import Validator
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword


class InterfaceParameterExists(CfnLintKeyword):
    """Check if Metadata Interface parameters exist"""

    id = "W4001"
    shortdesc = "Metadata Interface parameters exist"
    description = "Metadata Interface parameters actually exist"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-cloudformation-interface.html"
    tags = ["metadata"]

    valid_keys = ["ParameterGroups", "ParameterLabels"]

    def __init__(self) -> None:
        super().__init__()
        self.keywords = [
            "cfnParameter",
        ]

    def validate(self, validator: Validator, _, instance: Any, schema):
        for err in validator.descend(
            instance, schema={"enum": list(validator.context.parameters.keys())}
        ):
            err.rule = self
            yield err



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from datetime import datetime

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.lmbd.DeprecatedRuntimeCreate import DeprecatedRuntimeCreate


@pytest.fixture(scope="module")
def rule():
    rule = DeprecatedRuntimeCreate()
    yield rule


@pytest.mark.parametrize(
    "instance,date,expected",
    [
        (
            "nodejs16.x",
            datetime(2023, 12, 14),
            [],
        ),
        (
            "foo",  # not existent runtime
            datetime(2023, 12, 14),
            [],
        ),
        (
            {},  # wrong type
            datetime(2023, 12, 14),
            [],
        ),
        (
            "python3.7",
            datetime(2024, 1, 9),
            [
                ValidationError(
                    (
                        "Runtime 'python3.7' was deprecated on "
                        "'2023-12-04'. Creation was disabled on "
                        "'2024-01-09' and update on '2025-11-01'. "
                        "Please consider updating to 'python3.13'"
                    ),
                )
            ],
        ),
        (
            # will be caught by the update rule
            "python3.7",
            datetime(2025, 11, 1),
            [],
        ),
        (
            # will be caught by the update rule
            "nodejs",
            datetime(2016, 10, 31),
            [],
        ),
    ],
)
def test_lambda_runtime(instance, date, expected, rule, validator):
    rule.current_date = date
    errs = list(rule.validate(validator, "LambdaRuntime", instance, {}))
    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque
from unittest import mock

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.lmbd.SnapStartSupported import SnapStartSupported


@pytest.fixture(scope="module")
def rule():
    rule = SnapStartSupported()
    yield rule


@pytest.mark.parametrize(
    "name,instance,regions,add_child_rule,child_rule_called,expected",
    [
        (
            "SnapStart enabled on appropriate java version",
            {
                "Runtime": "java17",
                "SnapStart": {
                    "ApplyOn": "PublishedVersions",
                },
            },
            ["us-east-1"],
            True,
            False,
            [],
        ),
        (
            "SnapStart enabled with ref for runtime",
            {
                "Runtime": {"Ref": "Runtime"},
                "SnapStart": {
                    "ApplyOn": "PublishedVersions",
                },
            },
            ["us-east-1"],
            True,
            False,
            [],
        ),
        (
            "SnapStart not enabled in region that doesn't support it",
            {
                "Runtime": "java17",
                "SnapStart": {
                    "ApplyOn": "PublishedVersions",
                },
            },
            ["us-east-1", "foo-bar-1"],
            True,
            False,
            [
                ValidationError(
                    "'SnapStart' enabled functions are not supported in ['foo-bar-1']",
                    path=deque(["SnapStart", "ApplyOn"]),
                )
            ],
        ),
        (
            "SnapStart enabled for python3.12 error",
            {
                "Runtime": "python3.12",
                "SnapStart": {
                    "ApplyOn": "PublishedVersions",
                },
            },
            ["us-east-1"],
            True,
            False,
            [],
        ),
        (
            "SnapStart enabled for dotnet",
            {
                "Runtime": "dotnet8",
                "SnapStart": {
                    "ApplyOn": "PublishedVersions",
                },
            },
            ["us-east-1"],
            True,
            False,
            [],
        ),
        (
            "SnapStart enabled for go that isn't supported",
            {
                "Runtime": "go1.x",
                "SnapStart": {
                    "ApplyOn": "PublishedVersions",
                },
            },
            ["us-east-1"],
            True,
            False,
            [
                ValidationError(
                    "'go1.x' is not supported for 'SnapStart' enabled functions",
                    path=deque(["SnapStart", "ApplyOn"]),
                )
            ],
        ),
        (
            "SnapStart enabled for dotnet version that isn't supported",
            {
                "Runtime": "dotnet5.0",
                "SnapStart": {
                    "ApplyOn": "PublishedVersions",
                },
            },
            ["us-east-1"],
            True,
            False,
            [
                ValidationError(
                    "'dotnet5.0' is not supported for 'SnapStart' enabled functions",
                    path=deque(["SnapStart", "ApplyOn"]),
                )
            ],
        ),
        (
            "SnapStart enabled for dotnetcore version that isn't supported",
            {
                "Runtime": "dotnetcore2.1",
                "SnapStart": {
                    "ApplyOn": "PublishedVersions",
                },
            },
            ["us-east-1"],
            True,
            False,
            [
                ValidationError(
                    (
                        "'dotnetcore2.1' is not supported for "
                        "'SnapStart' enabled functions"
                    ),
                    path=deque(["SnapStart", "ApplyOn"]),
                )
            ],
        ),
        (
            "SnapStart not enabled on python non supported runtime",
            {
                "Runtime": "python3.11",
            },
            ["us-east-1"],
            True,
            True,
            [],
        ),
        (
            "SnapStart not enabled on python runtime in a bad region",
            {
                "Runtime": "python3.11",
            },
            ["foo-bar-1"],
            True,
            False,
            [],
        ),
        (
            "SnapStart not enabled and no child rule",
            {
                "Runtime": "java17",
            },
            ["us-east-1"],
            False,
            False,
            [],
        ),
        (
            "SnapStart set off with Python runtime",
            {
                "Runtime": "python3.11",
                "SnapStart": {
                    "ApplyOn": "None",
                },
            },
            ["us-east-1"],
            True,
            False,
            [],
        ),
        (
            "Snapstart should not be enabled for non java runtime",
            {
                "Runtime": "python3.11",
                "SnapStart": {
                    "ApplyOn": "PublishedVersions",
                },
            },
            ["us-east-1"],
            True,
            False,
            [
                ValidationError(
                    "'python3.11' is not supported for 'SnapStart' enabled functions",
                    path=deque(["SnapStart", "ApplyOn"]),
                )
            ],
        ),
    ],
)
def test_validate(
    name,
    instance,
    regions,
    add_child_rule,
    child_rule_called,
    expected,
    rule,
    validator,
):
    validator = validator.evolve(
        context=validator.context.evolve(regions=regions),
    )

    child_rule = mock.MagicMock()
    if add_child_rule:
        rule.child_rules["I2530"] = child_rule
    else:
        rule.child_rules["I2530"] = None

    errs = list(rule.validate(validator, "", instance, {}))

    if child_rule_called:
        child_rule.validate.assert_called_with(
            instance.get("Runtime")
        ), f"{name!r}: child rule not called"
    else:
        child_rule.validate.assert_not_called(), f"{name!r}: child rule called"

    assert errs == expected, f"{name!r}: expected {expected!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.functions.SubNotJoin import SubNotJoin


@pytest.fixture(scope="module")
def rule():
    rule = SubNotJoin()
    yield rule


@pytest.fixture
def template():
    return {
        "Parameters": {
            "MyList": {
                "Type": "CommaDelimitedList",
            },
        },
        "Resources": {
            "MyResource": {
                "Type": "AWS::S3::Bucket",
            },
        },
    }


@pytest.mark.parametrize(
    "name,instance,schema,expected",
    [
        (
            "Valid Fn::Join with proper delimiter",
            {"Fn::Join": [",", ["", ""]]},
            {"type": "string"},
            [],
        ),
        (
            "Valid Fn::Join with a list parameter",
            {"Fn::Join": ["", {"Ref": "MyList"}]},
            {"type": "string"},
            [],
        ),
        (
            "Invalid Fn::Join with an empty string",
            {"Fn::Join": ["", ["foo", "bar"]]},
            {"type": "string"},
            [
                ValidationError(
                    "Prefer using Fn::Sub over Fn::Join with an empty delimiter",
                    path=deque(["Fn::Join", 0]),
                    rule=SubNotJoin(),
                )
            ],
        ),
        (
            "Valid Fn::Join with a Fn::Sub",
            {"Fn::Join": ["", ["foo", {"Fn::Sub": ["MyResource", {}]}]]},
            {"type": "string"},
            [],
        ),
        (
            "Invalid Fn::Join with a Ref",
            {"Fn::Join": ["", ["foo", {"Ref": "MyResource"}]]},
            {"type": "string"},
            [
                ValidationError(
                    "Prefer using Fn::Sub over Fn::Join with an empty delimiter",
                    path=deque(["Fn::Join", 0]),
                    rule=SubNotJoin(),
                )
            ],
        ),
        (
            "Invalid Fn::Join with a Fn::Sub",
            {"Fn::Join": ["", ["foo", {"Fn::Sub": "${MyResource}"}]]},
            {"type": "string"},
            [
                ValidationError(
                    "Prefer using Fn::Sub over Fn::Join with an empty delimiter",
                    path=deque(["Fn::Join", 0]),
                    rule=SubNotJoin(),
                )
            ],
        ),
    ],
)
def test_validate(name, instance, schema, expected, rule, validator):
    errs = list(rule.validate(validator, schema, instance, {}))
    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from typing import Any

from cfnlint.jsonschema import ValidationError, Validator
from cfnlint.rules.functions._BaseFn import BaseFn


class DynamicReferenceSecureString(BaseFn):
    id = "E1027"
    shortdesc = "Check dynamic references secure strings are in supported locations"
    description = (
        "Dynamic References Secure Strings are only supported for a small set of"
        " resource properties.  Validate that they are being used in the correct"
        " location when checking values and Fn::Sub in resource properties. Currently"
        " doesn't check outputs, maps, conditions, parameters, and descriptions."
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/dynamic-references.html"
    tags = ["functions", "dynamic reference"]

    def __init__(self):
        """Init"""
        super().__init__()
        self.exceptions = [
            "Resources/AWS::DirectoryService::MicrosoftAD/Properties/Password",
            "Resources/AWS::DirectoryService::SimpleAD/Properties/Password",
            "Resources/AWS::ElastiCache::ReplicationGroup/Properties/AuthToken",
            "Resources/AWS::IAM::User/Properties/LoginProfile/Password",
            "Resources/AWS::KinesisFirehose::DeliveryStream/Properties/RedshiftDestinationConfiguration/Password",
            "Resources/AWS::OpsWorks::App/Properties/AppSource/Password",
            "Resources/AWS::OpsWorks::Stack/Properties/RdsDbInstances/*/DbPassword",
            "Resources/AWS::OpsWorks::Stack/Properties/CustomCookbooksSource/Password",
            "Resources/AWS::RDS::DBCluster/Properties/MasterUserPassword",
            "Resources/AWS::RDS::DBInstance/Properties/MasterUserPassword",
            "Resources/AWS::Redshift::Cluster/Properties/MasterUserPassword",
        ]

    def validate(self, validator: Validator, s: Any, instance: Any, schema: Any):
        if validator.context.path.cfn_path_string in self.exceptions:
            return

        yield ValidationError(
            (
                f"Dynamic reference {instance!r} to SSM secure strings "
                "can only be used in resource properties"
            ),
            rule=self,
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import logging
from collections import deque
from ipaddress import IPv4Network, IPv6Network, ip_network
from typing import Any

from cfnlint.context.conditions import Unsatisfiable
from cfnlint.helpers import ensure_list, is_function
from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.helpers import get_value_from_path
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword

LOGGER = logging.getLogger(__name__)


class VpcSubnetOverlap(CfnLintKeyword):
    id = "E3060"
    shortdesc = "Validate subnet CIDRs do not overlap with other subnets"
    description = (
        "When specifying subnet CIDRs for a VPC the subnet CIDRs "
        "most not overlap with eachother"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-ec2-subnet.html"
    tags = ["resources", "ec2", "vpc", "subnet"]

    def __init__(self) -> None:
        super().__init__(
            keywords=[
                "Resources/AWS::EC2::Subnet/Properties",
            ],
        )
        self._subnets: dict[str, list[tuple[IPv4Network | IPv6Network, dict]]] = {}

    def initialize(self, cfn):
        self._subnets = {}
        return super().initialize(cfn)

    def _validate_subnets(
        self,
        source: IPv4Network | IPv6Network,
        destination: IPv4Network | IPv6Network,
    ) -> bool:
        if isinstance(source, IPv4Network) and isinstance(destination, IPv4Network):
            if source.overlaps(destination):
                return True
            return False
        elif isinstance(source, IPv6Network) and isinstance(destination, IPv6Network):
            if source.overlaps(destination):
                return True
            return False
        return False

    def _create_network(self, cidr: Any) -> IPv4Network | IPv6Network | None:
        if not isinstance(cidr, str):
            return None

        try:
            return ip_network(cidr)
        except Exception as e:
            LOGGER.debug(f"Unable to create network from {cidr}", e)

        return None

    def validate(
        self, validator: Validator, keywords: Any, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:

        for vpc_id, vpc_validator in get_value_from_path(
            validator=validator, instance=instance, path=deque(["VpcId"])
        ):

            if not isinstance(vpc_id, (str, dict)):
                return

            fn_k, fn_v = is_function(vpc_id)
            if fn_k == "Fn::GetAtt":
                vpc_id = ensure_list(fn_v)[0].split(".")[0]
            elif fn_k == "Ref":
                vpc_id = fn_v
            elif fn_k:
                # its a function that we can't resolve
                return

            if not validator.is_type(vpc_id, "string"):
                return
            if vpc_id not in self._subnets:
                self._subnets[vpc_id] = []

            for key in ["CidrBlock", "Ipv6CidrBlock"]:
                for cidr_block, cidr_block_validator in get_value_from_path(
                    validator=vpc_validator, instance=instance, path=deque([key])
                ):

                    cidr_network = self._create_network(cidr_block)
                    if not cidr_network:
                        continue

                    for saved_subnet, conditions in self._subnets[vpc_id]:
                        # attempt to validate if the saved conditions comply
                        # with these conditions
                        try:
                            cidr_block_validator.evolve(
                                context=cidr_block_validator.context.evolve(
                                    conditions=cidr_block_validator.context.conditions.evolve(
                                        conditions
                                    )
                                )
                            )
                        except Unsatisfiable:
                            continue

                        # now we can evaluate if they overlap
                        if self._validate_subnets(
                            cidr_network,
                            saved_subnet,
                        ):
                            yield ValidationError(
                                (
                                    f"{str(cidr_network)!r} overlaps "
                                    f"with {str(saved_subnet)!r}"
                                ),
                                rule=self,
                                path=deque(
                                    list(cidr_block_validator.context.path.path)[1:]
                                ),
                            )

                    self._subnets[vpc_id].append(
                        (cidr_network, cidr_block_validator.context.conditions.status)
                    )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any

from sympy import And, Not, Or
from sympy.logic.boolalg import BooleanFunction

from cfnlint.conditions._utils import get_hash
from cfnlint.context.conditions._equals import Equal
from cfnlint.helpers import FUNCTION_CONDITIONS, is_function


@dataclass(frozen=True)
class Condition:
    instance: Any = field(init=True)
    status: bool | None = field(init=True, default=None)
    hash: str = field(init=False)

    fn_equals: Equal | None = field(init=True, default=None)
    condition: list["Condition"] | "Condition" | None = field(init=True, default=None)
    cnf: BooleanFunction = field(init=True, default_factory=BooleanFunction)

    def __post_init__(self):
        object.__setattr__(self, "hash", get_hash(self.instance))

    @classmethod
    def create_from_instance(
        cls, instance: Any, all_conditions: dict[str, Any]
    ) -> "Condition":
        fn_k, fn_v = is_function(instance)
        if fn_k is None:
            raise ValueError("Condition value must be an object of length 1")
        if fn_k in FUNCTION_CONDITIONS:
            if not isinstance(fn_v, list):
                raise ValueError(f"{fn_v!r} value should be an array")
            if fn_k == "Fn::Equals":
                equal = Equal.create_from_instance(fn_v)
                return cls(instance=instance, fn_equals=equal, cnf=equal.cnf)

            condition = []
            for v in fn_v:
                condition.append(Condition.create_from_instance(v, all_conditions))

            cnf = None
            if fn_k == "Fn::And":
                cnf = And(*[c.cnf for c in condition])
            elif fn_k == "Fn::Or":
                cnf = Or(*[c.cnf for c in condition])
            elif fn_k == "Fn::Not":
                if len(condition) != 1:
                    raise ValueError(
                        f"Fn::Not expects only one condition, got {len(condition)}"
                    )
                cnf = Not(condition[0].cnf)

            return cls(instance=instance, condition=condition, cnf=cnf)

        if fn_k == "Condition":
            if not isinstance(fn_v, str):
                raise ValueError(f"Condition value {fn_v!r} must be a string")
            sub_condition = all_conditions.get(fn_v)
            try:
                sub_all_conditions = all_conditions.copy()
                del sub_all_conditions[fn_v]
                c = Condition.create_from_instance(sub_condition, sub_all_conditions)
            except Exception:
                c = Condition.create_from_instance(
                    {"Fn::Equals": [None, None]}, all_conditions
                )
            return cls(instance=instance, condition=c, cnf=c.cnf)

        raise ValueError(f"Unknown key {fn_k!r} in condition")

    def evolve(self, status: bool | None) -> "Condition":
        cls = self.__class__

        if self.status is not None:
            if status != self.status:
                raise ValueError(f"Resetting status to {status} from {self.status}")

        return cls(
            instance=self.instance,
            status=status,
            cnf=self.cnf,
        )

    @property
    def is_region(self) -> bool:
        """Returns True or False if the condition is based on region

        Args: None

        Returns:
            bool
                Returns True or False if the condition is based on region
        """
        if self.fn_equals:
            return self.fn_equals.is_region
        if isinstance(self.condition, list):
            for c in self.condition:
                if c.is_region:
                    return True
            return False
        if self.condition:
            return self.condition.is_region
        return False

    @property
    def equals(self) -> list[Equal]:
        """Returns a Sequence of the Equals that make up the Condition

        Args: None

        Returns:
            Sequence[EqualParameter] | Sequence[Equal] | None:
                The Equal that are part of the condition
        """
        if self.fn_equals:
            return [self.fn_equals]
        if isinstance(self.condition, list):
            equals = []
            for c in self.condition:
                equals.extend(c.equals)
            return equals
        if self.condition:
            return self.condition.equals
        return []



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque
from typing import Iterable

import pytest

from cfnlint.context import Path
from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.codepipeline.PipelineArtifactNames import (
    PipelineArtifactNames,
)


@pytest.fixture
def rule():
    rule = PipelineArtifactNames()
    yield rule


@pytest.fixture
def template():
    return {
        "Conditions": {
            "IsUsEast1": {"Fn::Equals": [{"Ref": "AWS::Region"}, "us-east-1"]},
            "IsUsEast2": {"Fn::Equals": [{"Ref": "AWS::Region"}, "us-east-2"]},
            "IsUsEast": {
                "Fn::Or": [
                    {"Condition": "IsUsEast1"},
                    {"Condition": "IsUsEast2"},
                ]
            },
        }
    }


_standard_path = deque(["Resources", "TestPipeline", "Properties", "Stages"])


def _append_queues(queue1: Iterable, queue2: Iterable) -> deque:
    new_queue: deque = deque()
    new_queue.extend(queue1)
    new_queue.extend(queue2)
    return new_queue


@pytest.mark.parametrize(
    "instances,expected",
    [
        (
            [
                (
                    [],
                    _append_queues(
                        _standard_path, [0, "Actions", 0, "OutputArtifacts", 0, "Name"]
                    ),
                    {},
                ),
                (
                    [],
                    _standard_path,
                    {},
                ),
            ],
            [],
        ),
        (
            [
                (
                    "Foo",
                    _append_queues(
                        _standard_path, [0, "Actions", 0, "OutputArtifacts", 0, "Name"]
                    ),
                    {},
                ),
                (
                    "Bar",
                    _append_queues(
                        _standard_path, [0, "Actions", 0, "OutputArtifacts", 1, "Name"]
                    ),
                    {},
                ),
                (
                    "Foo",
                    _append_queues(
                        _standard_path, [0, "Actions", 1, "InputArtifacts", 0, "Name"]
                    ),
                    {},
                ),
                (
                    "Bar",
                    _append_queues(
                        _standard_path, [0, "Actions", 1, "InputArtifacts", 1, "Name"]
                    ),
                    {},
                ),
            ],
            [],
        ),
        (
            [
                (
                    "Foo",
                    _append_queues(
                        _standard_path, [0, "Actions", 0, "OutputArtifacts", 0, "Name"]
                    ),
                    {},
                ),
                (
                    "Foo",
                    deque(
                        [
                            "Resources",
                            1,
                            "Properties",
                            "Stages",
                            0,
                            "Actions",
                            0,
                            "OutputArtifacts",
                            0,
                            "Name",
                        ]
                    ),
                    {},
                ),
            ],
            [],
        ),
        (
            [
                (
                    "Foo",
                    _append_queues(
                        _standard_path, [0, "Actions", 0, "OutputArtifacts", 0, "Name"]
                    ),
                    {},
                ),
                (
                    "Foo",
                    deque(
                        [
                            "Resources",
                            "AnotherPipeline",
                            "Properties",
                            "Stages",
                            0,
                            "Actions",
                            0,
                            "OutputArtifacts",
                            0,
                            "Name",
                        ]
                    ),
                    {},
                ),
            ],
            [],
        ),
        (
            [
                (
                    "Foo",
                    _append_queues(
                        _standard_path, [0, "Actions", 0, "OutputArtifacts", 0, "Name"]
                    ),
                    {},
                ),
                (
                    "Foo",
                    _append_queues(
                        _standard_path, [0, "Actions", 0, "OutputArtifacts", 1, "Name"]
                    ),
                    {},
                ),
            ],
            [
                ValidationError(
                    "'Foo' is already a defined 'OutputArtifact' Name",
                    rule=PipelineArtifactNames(),
                )
            ],
        ),
        (
            [
                (
                    "Foo",
                    _append_queues(
                        _standard_path, [0, "Actions", 0, "OutputArtifacts", 0, "Name"]
                    ),
                    {
                        "IsUsEast1": True,
                        "IsUsEast2": False,
                    },
                ),
                (
                    "Foo",
                    _append_queues(
                        _standard_path, [0, "Actions", 1, "OutputArtifacts", 0, "Name"]
                    ),
                    {
                        "IsUsEast1": False,
                        "IsUsEast2": True,
                    },
                ),
                (
                    "Foo",
                    _append_queues(
                        _standard_path, [0, "Actions", 1, "InputArtifacts", 0, "Name"]
                    ),
                    {
                        "IsUsEast1": True,
                        "IsUsEast2": False,
                    },
                ),
                (
                    "Foo",
                    _append_queues(
                        _standard_path, [0, "Actions", 1, "InputArtifacts", 1, "Name"]
                    ),
                    {
                        "IsUsEast1": False,
                        "IsUsEast2": True,
                    },
                ),
            ],
            [],
        ),
        (
            [
                (
                    "Foo",
                    _append_queues(
                        _standard_path, [0, "Actions", 0, "OutputArtifacts", 0, "Name"]
                    ),
                    {
                        "IsUsEast1": True,
                        "IsUsEast2": False,
                    },
                ),
                (
                    "Foo",
                    _append_queues(
                        _standard_path, [0, "Actions", 1, "OutputArtifacts", 0, "Name"]
                    ),
                    {
                        "IsUsEast1": True,
                        "IsUsEast2": False,
                    },
                ),
                (
                    "Foo",
                    _append_queues(
                        _standard_path, [0, "Actions", 0, "InputArtifacts", 0, "Name"]
                    ),
                    {
                        "IsUsEast1": False,
                        "IsUsEast2": True,
                    },
                ),
            ],
            [
                ValidationError(
                    "'Foo' is already a defined 'OutputArtifact' Name",
                    rule=PipelineArtifactNames(),
                ),
                ValidationError(
                    "'Foo' is not previously defined as an 'OutputArtifact'",
                    rule=PipelineArtifactNames(),
                ),
            ],
        ),
    ],
)
def test_validate(instances, expected, rule, validator):

    errs = []
    for instance in instances:
        instance_validator = validator.evolve(
            context=validator.context.evolve(
                path=Path(path=instance[1]),
                conditions=validator.context.conditions.evolve(
                    instance[2],
                ),
            )
        )
        errs.extend(list(rule.validate(instance_validator, "", instance[0], {})))

    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import pytest

from cfnlint.rules.formats.IamRoleArn import IamRoleArn


@pytest.fixture(scope="module")
def rule():
    rule = IamRoleArn()
    yield rule


@pytest.mark.parametrize(
    "name,instance,expected",
    [
        (
            "Valid IAM Role arn",
            "arn:aws:iam::123456789012:role/test-role",
            True,
        ),
        (
            "Invalid Iam role Arn",
            "arn:partition:iam::123456789012:role/test-role",
            False,
        ),
    ],
)
def test_validate(name, instance, expected, rule, validator):
    result = rule.format(validator, instance)
    assert result == expected, f"Test {name!r} got {result!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from copy import deepcopy
from typing import Any

from cfnlint.helpers import is_custom_resource
from cfnlint.jsonschema import ValidationResult, Validator
from cfnlint.jsonschema._utils import Unset
from cfnlint.rules.formats._schema_comparer import compare_schemas
from cfnlint.rules.jsonschema import CfnLintKeyword
from cfnlint.schema import PROVIDER_SCHEMA_MANAGER


class RefFormat(CfnLintKeyword):
    id = "E1041"
    shortdesc = "Check if Ref matches destination format"
    description = (
        "When source and destination format exists validate that they match in a Ref"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/best-practices.html#parmtypes"
    tags = ["functions", "ref"]

    def __init__(self):
        super().__init__(["*"])
        self.parent_rules = ["E1020"]

    def _filter_schema(
        self, validator: Validator, type_name: str, id: str, schema: dict[str, Any]
    ) -> dict[str, Any]:
        if type_name != "AWS::EC2::SecurityGroup":
            return schema

        items = list(
            validator.cfn.get_cfn_path(
                ["Resources", id, "Properties", "VpcId"], validator.context
            )
        )
        if items:
            # VpcId is specified and will have a value which means the returned value is
            # "AWS::EC2::SecurityGroup.Id"
            schema = deepcopy(schema)
            schema.pop("anyOf")
            schema["format"] = "AWS::EC2::SecurityGroup.Id"
            return schema

        # VpcId being None means it wasn't specified and the value is a Name
        schema = deepcopy(schema)
        schema.pop("anyOf")
        schema["format"] = "AWS::EC2::SecurityGroup.Name"
        return schema

    def validate(
        self, validator: Validator, _, instance: Any, schema: Any
    ) -> ValidationResult:
        if instance in validator.context.parameters:
            return

        if instance not in validator.context.resources:
            return
        t = validator.context.resources[instance].type

        # When using a custom resource you can set the physical ID
        # which results in a usable Ref
        if is_custom_resource(t):
            return

        for (
            regions,
            resource_schema,
        ) in PROVIDER_SCHEMA_MANAGER.get_resource_schemas_by_regions(
            t, validator.context.regions
        ):
            region = regions[0]
            ref_schema = validator.context.resources[instance].ref(region)
            ref_schema = self._filter_schema(validator, t, instance, ref_schema)

            err = compare_schemas(schema, ref_schema)
            if err:
                if err.instance:
                    err.message = (
                        f"{{'Ref': {instance!r}}} with formats {err.instance!r} "
                        "does not match destination format of "
                        f"{err.schema.get('format')!r}"
                    )
                else:
                    err.message = (
                        f"{{'Ref': {instance!r}}} does not match "
                        f"destination format of {err.schema.get('format')!r}"
                    )

                err.instance = Unset()
                err.schema = Unset()
                err.rule = self
                yield err



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque
from typing import Any

from cfnlint.helpers import REGEX_DYN_REF
from cfnlint.jsonschema import ValidationError, Validator
from cfnlint.rules.functions._BaseFn import BaseFn

_all = {
    "prefixItems": [
        {"type": "string", "const": "resolve"},
        {"type": "string", "enum": ["ssm", "ssm-secure", "secretsmanager"]},
    ],
    "minItems": 3,
}

_ssm = {
    "prefixItems": [
        {"type": "string", "const": "resolve"},
        {"type": "string", "enum": ["ssm", "ssm-secure"]},
        {"type": "string", "pattern": "[a-zA-Z0-9_.-/]+"},
        {"type": "string", "pattern": "\\d+"},
    ],
    "maxItems": 4,
    "type": "array",
}

_secrets_manager = {
    "prefixItems": [
        {"type": "string", "const": "resolve"},
        {"type": "string", "const": "secretsmanager"},
        {"type": "string", "pattern": "[ -~]*"},  # secret-id
        {"type": "string", "enum": ["SecretString", ""]},  # secret-string
        {"type": "string"},  # json-key
        {"type": "string"},  # version-stage or version-id
        {"type": "string"},  # version-stage or version-id
    ],
    "minItems": 3,
    "maxItems": 7,
    "type": "array",
}

_secrets_manager_arn = {
    "prefixItems": [
        {"type": "string", "const": "resolve"},
        {"type": "string", "const": "secretsmanager"},
        {"type": "string", "const": "arn"},  # arn
        {"type": "string"},  # partition
        {"type": "string"},  # service
        {"type": "string"},  # region
        {"type": "string"},  # account ID
        {"type": "string", "const": "secret"},  # secret
        {"type": "string"},  # secret
        {"type": "string", "enum": ["SecretString", ""]},  # secret-string
        {"type": "string"},  # json-key
        {"type": "string"},  # version-stage or version-id
        {"type": "string"},  # version-stage or version-id
    ],
    "minItems": 9,
    "maxItems": 13,
    "type": "array",
}


class DynamicReference(BaseFn):
    id = "E1050"
    shortdesc = "Validate the structure of a dynamic reference"
    description = "Make sure dynamic reference strings have the correct syntax"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/dynamic-references.html"
    tags = ["functions", "dynamic reference"]

    def __init__(self) -> None:
        super().__init__()
        self.child_rules = {
            "E1051": None,
            "W1051": None,
            "E1027": None,
        }

    def _clean_errors(self, err: ValidationError) -> ValidationError:
        err.rule = self
        err.path = deque([])
        err.schema_path = deque([])
        return err

    def dynamicReference(
        self, validator: Validator, s: Any, instance: Any, schema: Any
    ):
        if not validator.is_type(instance, "string"):
            return

        # SSM parameters can be used in Resources and Outputs and Parameters
        # SSM secrets are only used in a small number of locations
        # Secrets manager can be used only in resource properties
        validator = validator.evolve(
            function_filter=validator.function_filter.evolve(
                add_cfn_lint_keyword=False,
            ),
        )

        for v in REGEX_DYN_REF.findall(instance):
            parts = v.split(":")

            found = False
            evolved = validator.evolve(schema=_all)
            for err in evolved.iter_errors(parts):
                yield self._clean_errors(err)
                found = True

            if found:
                continue

            if parts[1] == "ssm":
                evolved = validator.evolve(schema=_ssm)
            elif parts[1] == "ssm-secure":
                evolved = validator.evolve(schema=_ssm)
                rule = self.child_rules["E1027"]
                if rule and hasattr(rule, "validate"):
                    yield from rule.validate(validator, {}, v, schema)
            else:
                if parts[2] == "arn":
                    evolved = validator.evolve(schema=_secrets_manager_arn)
                else:  # this is secrets manager
                    evolved = validator.evolve(schema=_secrets_manager)
                for rule_id in ["E1051", "W1051"]:
                    rule = self.child_rules[rule_id]
                    if rule and hasattr(rule, "validate"):
                        yield from rule.validate(validator, {}, v, schema)

            for err in evolved.iter_errors(parts):
                yield self._clean_errors(err)
